from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

import pytest

from axtreme.eval.object_logging import NestedDict, unpack_object, unpack_object_str_content


@dataclass
class Level2:
    a: str = "blah"
    b: None = None
    c: tuple[int, ...] = (1, 2, 3)


@dataclass
class Level1:
    level2: Level2
    x: str = "x"


@pytest.fixture
def l1():
    l2 = Level2()
    return Level1(level2=l2)


def test_unpack_object_depth_0(l1: Level1):
    unpacked = unpack_object(l1, depth=0)
    assert unpacked == {"__class__": Level1}


def test_unpack_object_depth_1(l1: Level1):
    unpacked = unpack_object(l1, depth=1)

    # fmt: off
    expected_value = {
        "__class__": Level1,
        "level2": {
            "__class__": Level2
        },
        "x": "x"
    }
    # fmt: on

    assert unpacked == expected_value


def test_unpack_object_depth_2(l1: Level1):
    unpacked = unpack_object(l1, depth=2)

    # fmt: off
    expected_value = {
        '__class__': Level1,
        'level2': {
            '__class__': Level2,
            'a': 'blah',
            'b': None,
            'c': (1, 2, 3)
        },
        'x': 'x'
    }
    # fmt: on

    assert unpacked == expected_value


def test_unpack_object_custom_config_depth_0(l1: Level1):
    """Custom config will be found ar depth 0 and override the entire result."""
    config: dict[type, Callable[[Any], NestedDict]] = {Level1: lambda _: {"junk1": "junk1"}}
    unpacked = unpack_object(l1, custom_unpacking_config=config, depth=2)
    assert unpacked == {"junk1": "junk1"}


def test_unpack_object_custom_config_depth_1(l1: Level1):
    """Custom config will overide the portion of the result"""
    config: dict[type, Callable[[Any], NestedDict]] = {Level2: lambda _: {"junk1": "junk1"}}
    unpacked = unpack_object(l1, custom_unpacking_config=config, depth=2)

    # fmt: off
    expected_value = {
        "__class__": Level1,
        "level2": {
            "junk1": "junk1"
        },
        "x": "x"
    }
    # fmt: on
    assert unpacked == expected_value


def test_unpack_object_no__dict__method():
    unpacked = unpack_object([1, 2, 3], depth=1)

    assert unpacked == {"__class__": list}


def test_unpack_object_str_content(l1: Level1):
    unpacked = unpack_object_str_content(l1, depth=2)

    # fmt: off
    expected_value = {
        '__class__': str(Level1),
        'level2': {
            '__class__': str(Level2),
            'a': 'blah',
            'b': str(None),
            'c': str((1, 2, 3))
        },
        'x': 'x'
    }
    # fmt: on

    assert unpacked == expected_value


@pytest.mark.integration
def test_unpack_object_integration():
    """This is mainly to document/demonstrate its functionality on more complicated objects."""

    from collections.abc import Sized
    from typing import cast

    import numpy as np
    import torch
    from torch.utils.data import DataLoader

    from axtreme.data import BatchInvariantSampler2d, MinimalDataset, SizableSequentialSampler
    from axtreme.qoi import GPBruteForce
    from axtreme.sampling import NormalIndependentSampler

    torch.set_default_dtype(torch.float64)

    N_ENV_SAMPLES_PER_PERIOD = 31  # noqa: N806

    dataset = MinimalDataset(np.arange(10_000).reshape(-1, 2))
    posterior_sampler = NormalIndependentSampler(torch.Size([50]))
    # gp brute force
    n_periods = 50
    sampler = SizableSequentialSampler(
        data_source=cast(Sized, dataset),
        num_samples=N_ENV_SAMPLES_PER_PERIOD * n_periods,
    )
    batch_sampler = BatchInvariantSampler2d(
        sampler=sampler,
        batch_shape=torch.Size([n_periods, 64]),
    )
    dataloader = DataLoader(dataset, batch_sampler=batch_sampler)

    qoi_estimator_gp_brute_force = GPBruteForce(
        env_iterable=dataloader,
        posterior_sampler=posterior_sampler,
        no_grad=True,
        seed=1234,
    )

    # here we simply check that it can run to completion. It will produce result similar to the ones below.
    # we don't check the actual results as this will change if underling methods add attributes etc.
    # That is too much overhead to keep this test up to date.
    _ = unpack_object_str_content(qoi_estimator_gp_brute_force, depth=2)
    # Expected output:
    # fmt: off
    _ = {
        "__class__": "<class 'axtreme.qoi.gp_bruteforce.GPBruteForce'>",
        "device": "cpu",
        "env_iterable": {
            "__class__": "<class 'torch.utils.data.dataloader.DataLoader'>",
            "batch_sampler": {
                "__class__": "<class 'axtreme.data.batch_invariant_sampler.BatchInvariantSampler2d'>"
            },
            "batch_size": "None",
            "collate_fn": {
                "__class__": "<class 'function'>"
            },
            "dataset": {
                "__class__": "<class 'axtreme.data.dataset.MinimalDataset'>"
            },
            "drop_last": "False",
            "generator": "None",
            "num_workers": "0",
            "persistent_workers": "False",
            "pin_memory": "False",
            "pin_memory_device": "",
            "prefetch_factor": "None",
            "sampler": {
                "__class__": "<class 'torch.utils.data.sampler.SequentialSampler'>"
            },
            "timeout": "0",
            "worker_init_fn": "None"
        },
        "erd_samples_per_period": "1",
        "input_transform": "None",
        "no_grad": "True",
        "outcome_transform": "None",
        "posterior_sampler": {
            "__class__": "<class 'axtreme.sampling.normal_independent_sampler.NormalIndependentSampler'>",
            "sample_shape": "torch.Size([50])",
            "seed": "332061",
            "training": "True"
        },
        "seed": "1234"
    }
    # fmt: on



import numpy as np
import torch
from torch.utils.data import DataLoader

from axtreme.data import ImportanceAddedWrapper, ImportanceIndexWrapper, MinimalDataset


def test_ImportanceAddedWrapper_numpy_input():
    dataset = MinimalDataset(
        np.array(
            [
                [0, 1, 2],
                [3, 4, 5],
            ]
        )
    )
    index_dataset = ImportanceIndexWrapper(dataset, importance_idx=2)

    # run test
    actual_data, actual_weight = index_dataset[0]

    # perform testing:
    assert list(actual_data) == [0, 1]
    assert actual_weight == 2


def test_ImportanceAddedWrapper_torch():
    dataset = MinimalDataset(
        torch.tensor(
            [
                [0, 1, 2],
                [3, 4, 5],
            ]
        )
    )
    index_dataset = ImportanceIndexWrapper(dataset, importance_idx=2)

    # run test
    actual_data, actual_weight = index_dataset[0]

    # perform testing:
    assert actual_data.tolist() == [0, 1]
    assert actual_weight == 2


def test_ImportanceAddedWrapper():
    data = np.array(
        [
            [0, 1],
            [3, 4],
        ]
    )
    data_dataset = MinimalDataset(data)
    weights = np.array([2, 5])
    importance_datasets = MinimalDataset(weights)

    dataset = ImportanceAddedWrapper(data_dataset, importance_datasets)

    actual_data, actual_weight = dataset[0]

    assert list(actual_data) == [0, 1]
    assert actual_weight == 2


##### Intergration tests
# Primarily to make sure the weight are 1d
def test_ImportanceAddedWrapper_dataloader_intergration():
    data = np.array(
        [
            [0, 1],
            [3, 4],
        ],
        dtype=np.int32,
    )
    data_dataset = MinimalDataset(data)
    weights = np.array([2, 5], dtype=np.int32)
    importance_datasets = MinimalDataset(weights)

    dataset = ImportanceAddedWrapper(data_dataset, importance_datasets)

    dl = DataLoader(dataset, shuffle=False, batch_size=2)

    actual_data, actual_weight = next(iter(dl))

    torch.testing.assert_close(
        actual_data,
        torch.tensor(
            [
                [0, 1],
                [3, 4],
            ],
            dtype=torch.int32,
        ),
    )

    torch.testing.assert_close(
        actual_weight,
        torch.tensor(
            [2, 5],
            dtype=torch.int32,
        ),
    )



"""Mainly provides the different posterior to test sampling against.

Assume covariance (off diagonals) is 0 unless function mentions otherwise

These are defined as functions rather than fixtures as we often refere to them with `@pytest.mark.paramterise`.
Pytest does not allow you to refere to fixtures in parameterise. For discussion of alternative see
`docs/dev/design_decisons/pytest.md`.

NOTE: This uses the posterior dimension convention defined [here](https://docs.gpytorch.ai/en/stable/distributions.html#multitaskmultivariatenormal)
NOTE: This could be within conftest, but appears it is considerd bad practice to import from conftest https://github.com/pytest-dev/pytest/issues/3272
"""

# %%
import torch
from botorch.posteriors.gpytorch import GPyTorchPosterior
from gpytorch.distributions.multitask_multivariate_normal import MultitaskMultivariateNormal
from gpytorch.distributions.multivariate_normal import MultivariateNormal


def posterior_n1_t1() -> GPyTorchPosterior:
    """Create a posterior (represents gp prediction output).

    Represents:
        - 1 input point
        - 1d output dimension

    Resultant shape:
    >>> posterior.distribution.mean
    tensor([1])
    >>> posterior.distribution.covariance_matrix
    tensor([[1.]])

    Why test this case:
        - most basic example of a prediction
    """
    return GPyTorchPosterior(MultivariateNormal(mean=torch.tensor([1.0]), covariance_matrix=torch.eye(1)))


def posterior_n2_t1() -> GPyTorchPosterior:
    """Create a posterior (represents gp prediction output).

    Represents:
        - 2 input point.
        - 1d output dimension.
        - No covariance.

    Resultant shape:
    >>> posterior.distribution.mean
    tensor([1,1])
    >>> posterior.distribution.covariance_matrix
    tensor([[1., 0.],
            [0., 1.]]))

    Why test this case:
        - Show how shape extends with additional inputs compared to `posterior_1point_1Doutput`
    """
    return GPyTorchPosterior(MultivariateNormal(mean=torch.tensor([1.0, 1]), covariance_matrix=torch.eye(2)))


def posterior_n1_t2() -> GPyTorchPosterior:
    """Create a posterior (represents gp prediction output).

    Represents:
        - 1 input point.
        - 2d output dimension.
        - Independat targets
        - Within targets, independant x points.

    Resultant shape:
    >>> posterior.distribution.mean
    tensor([[1.0000, 0.5000]])
    >>> posterior.distribution.covariance_matrix
    tensor([[1.0000, 0.0000],
            [0.0000, 0.5000]]

    Why test this case:
        - Basic example of multi dimensional output space
    """
    mvn_task1 = MultivariateNormal(mean=torch.tensor([1.0]), covariance_matrix=torch.eye(1))
    mvn_task2 = MultivariateNormal(mean=torch.tensor([0.5]), covariance_matrix=torch.eye(1) * 0.5)
    return GPyTorchPosterior(MultitaskMultivariateNormal.from_independent_mvns([mvn_task1, mvn_task2]))


def posterior_n3_t2() -> GPyTorchPosterior:
    """Create a posterior (represents gp prediction output).

    Represents:
        - 3 input point.
        - 2d output dimension.
        - Independat targets
        - Within targets, independant x points.

    Resultant shape:
    >>> posterior.distribution.mean
    tensor([[1.0000, 0.5000],
            [1.0000, 0.5000],
            [1.0000, 0.5000]])
    >>> posterior.distribution.covariance_matrix
    tensor([[1.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.0000],
            [0.0000, 1.0000, 0.0000, 0.0000, 0.0000, 0.0000],
            [0.0000, 0.0000, 1.0000, 0.0000, 0.0000, 0.0000],
            [0.0000, 0.0000, 0.0000, 0.5000, 0.0000, 0.0000],
            [0.0000, 0.0000, 0.0000, 0.0000, 0.5000, 0.0000],
            [0.0000, 0.0000, 0.0000, 0.0000, 0.0000, 0.5000]]))

    Why test this case:
        - Show how shape extends with additional inputs compared to `posterior_1point_2Doutput`
    """
    mvn_task1 = MultivariateNormal(mean=torch.tensor([1.0, 1, 1]), covariance_matrix=torch.eye(3))
    mvn_task2 = MultivariateNormal(mean=torch.tensor([0.5, 5, 5]), covariance_matrix=torch.eye(3) * 0.5)
    return GPyTorchPosterior(MultitaskMultivariateNormal.from_independent_mvns([mvn_task1, mvn_task2]))


def posterior_n3_t2_interleaved() -> GPyTorchPosterior:
    """Creates a Multitask Multivariate Normal with interleaved tasks.

    In interleaved format the covariance matrix groups all targets for the same x point together. The mean is always
    (b,n,m). See example below.

        Non-interleaved format:
            x_1, ..., x_n, x_1, ..., x_n,
        x_1|              |               |
        ...|   Task 1     |   Cov Task1   |
        x_n|              |   & Task 2    |
        - |--------------|---------------|
        x_1|              |               |
        ...|   Cov Task1  |     Task 2    |
        x_n|   & Task 2   |               |
        --------------------------------

        Interleaved format:
            x_1, x1,..., ..., x_n, x_n,
        x_1| t1,
        x_1|   , t2
        ...|
        ...|
        x_n|
        x_n|

    Example:

        non_interleaved_cov = torch.tensor([
            [1.0,  0.0,  0.0,  0.0,  0.0,  0.0],
            [0.0,  2.0,  0.0,  0.0,  0.0,  0.0],
            [0.0,  0.0,  3.0,  0.0,  0.0,  0.0],
            [0.0,  0.0,  0.0, 11.0,  0.0,  0.0],
            [0.0,  0.0,  0.0,  0.0, 12.0,  0.0],
            [0.0,  0.0,  0.0,  0.0,  0.0, 13.0],
        ])

        interleaved_cov = torch.tensor([
            [1.0, 0.0,  0.0,  0.0,  0.0,  0.0],
            [0.0, 11.0, 0.0,  0.0,  0.0,  0.0],
            [0.0,  0.0, 2.0,  0.0,  0.0,  0.0],
            [0.0,  0.0, 0.0, 12.0,  0.0,  0.0],
            [0.0,  0.0, 0.0,  0.0,  3.0,  0.0],
            [0.0,  0.0, 0.0,  0.0,  0.0, 13.0],
        ],


    Why test this case:
        - MultitaskMultivariateNormal: Support both formats. Should ensure both are handled.
    """

    # fmt: off
    # Shape (*b,n,m)
    mean = torch.tensor(
        [
            [1 ,100],
            [2, 200],
            [3, 300],
        ]
    )

    cov_interleaved = torch.tensor(
        [
            [1.0, 0.0,  0.0,  0.0,  0.0,  0.0],
            [0.0, 11.0, 0.0,  0.0,  0.0,  0.0],
            [0.0,  0.0, 2.0,  0.0,  0.0,  0.0],
            [0.0,  0.0, 0.0, 12.0,  0.0,  0.0],
            [0.0,  0.0, 0.0,  0.0,  3.0,  0.0],
            [0.0,  0.0, 0.0,  0.0,  0.0, 13.0],
        ],
    )
    # fmt: on

    return GPyTorchPosterior(
        MultitaskMultivariateNormal(mean=mean, covariance_matrix=cov_interleaved, interleaved=True)
    )


def posterior_b2_n3_t2() -> GPyTorchPosterior:
    """Create a posterior (represents gp prediction output).

    Represents:
        - 2 batches
        - 3 input point.
        - 2d output dimension.
        - Independat targets
        - Within targets, independant x points.

    Resultant shape:
    >>> posterior.distribution.mean.shape
    torch.Size([2,3,2])
    >>> posterior.distribution.covariance_matrix.shape
    torch.Size([2,3*2,3*2])


    Why test this case:
        - Add posterior batch shape to `posterior_3point_2Doutput`.
    """
    # fmt: off
    batch_mean = torch.tensor(
        [[[1., 1.5],
          [1., 1.5],
          [1., 1.5]],
         [[2., 2.5],
          [2., 2.5],
          [2., 2.5]]]
    )
    # fmt: on

    batch_covariance = torch.tensor(
        [
            [
                [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 1.5, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.5, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.5],
            ],
            [
                [2.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 2.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 2.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 2.5, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 2.5, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 2.5],
            ],
        ]
    )

    return GPyTorchPosterior(
        MultitaskMultivariateNormal(mean=batch_mean, covariance_matrix=batch_covariance, interleaved=False)
    )


# %%
if __name__ == "__main__":
    """The following can be used to generate more posteriors should they be needed."""

    import torch
    from botorch.fit import fit_gpytorch_mll
    from botorch.models import SingleTaskGP
    from gpytorch.mlls import ExactMarginalLogLikelihood

    torch.set_default_dtype(torch.float64)

    def standardize_data(data: torch.Tensor):
        mean = data.mean(dim=0)
        std = data.std(dim=0)
        return (data - mean) / std

    def max_min_normalize_data(data: torch.Tensor):
        max_val = data.max(dim=0).values
        min_val = data.min(dim=0).values
        return (data - min_val) / (max_val - min_val)

    def make_posterior(n_test_points: int, y_dim: int, n_training_points: int = 10):
        # x dim is not included as it doesn't not effect the posterior dim
        train_x = max_min_normalize_data(torch.rand([n_training_points, 1]))
        train_y = standardize_data(torch.rand(n_training_points, y_dim))

        model = SingleTaskGP(train_x, train_y)
        mll = ExactMarginalLogLikelihood(model.likelihood, model)
        _ = fit_gpytorch_mll(mll)
        test_x = torch.rand([n_test_points, 1])
        with torch.no_grad():
            posterior = model.posterior(test_x)

        return posterior

    # posterior = make_posterior(n_test_points=2, y_dim=2)
    n_test_points = 3
    y_dim = 2
    n_training_points = 10
    # x dim is not included as it doesn't not effect the posterior dim
    train_x = max_min_normalize_data(torch.rand([2, n_training_points, 1]))
    train_y = standardize_data(torch.rand(2, n_training_points, y_dim))

    model = SingleTaskGP(train_x, train_y)
    mll = ExactMarginalLogLikelihood(model.likelihood, model)
    _ = fit_gpytorch_mll(mll)
    test_x = torch.rand([n_test_points, 1])
    with torch.no_grad():
        posterior = model.posterior(test_x)

    assert isinstance(posterior, GPyTorchPosterior)

    dist = posterior.distribution
    print(f"{dist.mean.shape= }, {dist.covariance_matrix.shape= }")  # noqa: T201
    # print(f"{type(dist)}\n{dist.mean}\n{dist.covariance_matrix}")

    # %%
    dist = posterior_b2_n3_t2()
    print(f"{dist.mean.shape= }, {dist.covariance_matrix.shape= }")  # noqa: T201

    # %%
    dist.shape()
    # %%
    dist  # noqa: B018  # pyright: ignore[reportUnusedExpression]
    # %%



# sourcery skip: dont-import-test-modules

import logging
import os
import warnings
from pathlib import Path
from shutil import rmtree

import gpytorch
import matplotlib.pyplot as plt
import pytest
import torch
import torch.cuda
from botorch.fit import fit_gpytorch_mll
from botorch.models import SingleTaskGP

from axtreme.plotting.gp_fit import plot_1d_model
from tests import helpers

CUDA_AVAILABLE: bool = torch.cuda.is_available()

TORCH_DEVICES: list[str] = ["cuda", "cpu"] if CUDA_AVAILABLE else ["cpu"]

# gpytorch recommends opperating in float64
torch.set_default_dtype(torch.float64)


@pytest.fixture(scope="package", autouse=True)
def chdir() -> None:
    """
    Fixture that changes the current working directory to the 'test_working_directory' folder.
    This fixture is automatically used for the entire package.
    """
    os.chdir(Path(__file__).parent.absolute() / "test_working_directory")


# @pytest.fixture(scope="class", autouse=True)
# def reset_torch_default_device() -> None:
#     torch.set_default_device("cpu")


@pytest.fixture(scope="class", params=TORCH_DEVICES)
def vary_torch_default_device(request: pytest.FixtureRequest):
    torch.set_default_device(request.param)
    yield
    torch.set_default_device("cpu")  # reset to default device after test


@pytest.fixture(scope="package", autouse=True)
def test_dir() -> Path:
    """
    Fixture that returns the absolute path of the directory containing the current file.
    This fixture is automatically used for the entire package.
    """
    return Path(__file__).parent.absolute()


output_dirs = [
    "results",
    "data",
]
output_files = [
    "*test*.pdf",
]


@pytest.fixture(autouse=True)
def default_setup_and_teardown():
    """
    Fixture that performs setup and teardown actions before and after each test function.
    It removes the output directories and files specified in 'output_dirs' and 'output_files' lists.
    """
    _remove_output_dirs_and_files()
    yield
    _remove_output_dirs_and_files()


def _remove_output_dirs_and_files() -> None:
    """
    Helper function that removes the output directories and files specified in 'output_dirs' and 'output_files' lists.
    """
    for folder in output_dirs:
        rmtree(folder, ignore_errors=True)
    for pattern in output_files:
        for file in Path.cwd().glob(pattern):
            _file = Path(file)
            _file.unlink(missing_ok=True)


@pytest.fixture(autouse=True)
def setup_logging(caplog: pytest.LogCaptureFixture) -> None:
    """
    Fixture that sets up logging for each test function.
    It sets the log level to 'INFO' and clears the log capture.
    """
    caplog.set_level("INFO")
    caplog.clear()


@pytest.fixture(autouse=True)
def logger() -> logging.Logger:
    """Fixture that returns the logger object."""
    return logging.getLogger()


@pytest.fixture(
    params=[
        helpers.single_task_homo_noise_m_1(),
        helpers.single_task_homo_noise_m_2(),
        helpers.single_task_fixed_noise_m_1(),
        helpers.single_task_fixed_noise_m_2(),
        helpers.single_task_homo_noise_m_2_b3(),
    ],
    # provides naming to easily identify model used
    ids=[
        "single_task_homo_noise_m_1",
        "single_task_homo_noise_m_2",
        "single_task_fixed_noise_m_1",
        "single_task_fixed_noise_m_2",
        "single_task_homo_noise_m_2_b3",
    ],
)
def botorch_models(request: pytest.FixtureRequest):
    """Provides a generic set of model to be tested against.

    This is useful when functionality should be invariant to specific type of model.

    NOTE: Its not possible to paramterise fixture so we do this instead.
    """
    return request.param


@pytest.fixture(scope="session")
def model_singletaskgp_d1_t1(*, visual_inspect: bool = False):
    """Creates a basic gp with 1d input and 1d output."""

    def y_func(x: torch.Tensor) -> torch.Tensor:
        return torch.sin(x * 10)

    train_x = torch.tensor([[0.1], [0.5], [0.9], [1.0]])
    train_y = y_func(train_x)

    # Ignore the warning about centralise data, we have inspected it and are fine with it.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message=".*Please consider scaling the input.*")

        model = SingleTaskGP(train_X=train_x, train_Y=train_y, train_Yvar=torch.ones_like(train_y) * 0.1)
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(model.likelihood, model)
        _ = fit_gpytorch_mll(mll)

    if visual_inspect:
        _ = plot_1d_model(model)

    return model


@pytest.fixture(scope="session")
def model_singletaskgp_d1_t2(*, visual_inspect: bool = False):
    """Creates a basic gp with 1d input and 2d output.

    This output can be used to represent location and scale modeling. The functions have not been choosen in any
    particulat way, except the 2nd function (scale) is larget than 0.
    """

    def target1(x: torch.Tensor) -> torch.Tensor:
        return torch.sin(x * 10)

    def target2(x: torch.Tensor) -> torch.Tensor:
        return 0.5 * torch.cos(x * 10) + 0.8

    def y_func(x: torch.Tensor) -> torch.Tensor:
        return torch.concat([target1(x), target2(x)], dim=-1)

    train_x = torch.tensor([[0.1], [0.5], [0.9], [1.0]])
    train_y = y_func(train_x)

    # Ignore the warning about centralise data, we have inspected it and are fine with it.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message=".*Please consider scaling the input.*")

        model = SingleTaskGP(train_X=train_x, train_Y=train_y, train_Yvar=torch.ones_like(train_y) * 0.1)
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(model.likelihood, model)
        _ = fit_gpytorch_mll(mll)

    if visual_inspect:
        _, ax = plt.subplots(1, 1, figsize=(6, 4))
        x = torch.linspace(0, 1, 100).reshape(-1, 1)
        _ = ax.plot(x, target1(x), label="target1")
        _ = ax.plot(x, target2(x), label="target2")
        _ = plot_1d_model(model, ax=ax)
        _ = ax.legend()

    return model



import pytest
import torch
from botorch.models.deterministic import GenericDeterministicModel
from botorch.sampling.index_sampler import IndexSampler
from tests.qoi.utils import dummy_posterior_mean


@pytest.fixture
def gp_passthrough_1p() -> GenericDeterministicModel:
    """Simple Gp for testing, contain 1 unique posterior sample.

    Creates a deterministic gp which always produce indentical posterior samples.
    The output loc is a direct pass through of the input data, and the scale is set to 1e-6.

    See dummpy_posterior_mean for details.
    """
    return GenericDeterministicModel(dummy_posterior_mean, num_outputs=2)


@pytest.fixture
def gp_passthrough_1p_sampler() -> IndexSampler:
    """Default sampler to use with gp_passthrough_1p."""
    return IndexSampler(torch.Size([1]))



"""MarginalCDFExtrapolation calculates QoIs by estimating the behaviour of a single timestep, and extrapolating."""

# pyright: reportUnnecessaryTypeIgnoreComment=false
# %%
import warnings
from collections.abc import Iterable
from contextlib import ExitStack

import gpytorch
import torch
from botorch.models.model import Model
from botorch.models.transforms.input import InputTransform
from botorch.models.transforms.outcome import OutcomeTransform
from torch.distributions import Categorical, Distribution, Gumbel
from torch.nn.modules import Module

from axtreme.distributions import icdf
from axtreme.distributions.mixture import ApproximateMixture, icdf_value_bounds
from axtreme.qoi.qoi_estimator import QoIEstimator
from axtreme.sampling import MeanVarPosteriorSampledEstimates, PosteriorSampler, UTSampler


class MarginalCDFExtrapolation(MeanVarPosteriorSampledEstimates, QoIEstimator):
    """Estimate an ERD QoI using the marginal CDF for a single timestep.

    Basic Idea:
        - Get the marginal CDF of the response for a single timestep (e.g. 1 hour).
            Marginal CDF: The CDF after marginalising out all sources of randomness within a timestep. E.g if you
            considered all the variablity in a single timestep (e.g all the different weather conditions), collected
            the different CDFs they produce, and then averaged them, you would have the marginal CDF.
        - Using the average CDF you can then calculate:
            - "The probablity that a response of size `y` won't be exceeded in 1 hour" as follows:
                - :math:`CDF(y = .8) = .5`
            - "The probablity that a response of size `y` won't be exceeded in 2 hours" as follows:
                - ``[Prb not exceeded in hour one]`` AND ``[Prb not exceeded in hour two]``
                - :math:`CDF(y = .8) * CDF(y = .8) = .5 * .5 = .25`
            - "The probablity that a response of size `y` won't be exceeded in N hours" as follows:
                - :math:`CDF(y = .8)^N = (.5)^N`

        This is possible because we are using the 'average' timestep, which is the same for all timesteps.

    Strengths:
        - Once the 'average' CDF has been obtained, very large values of N can be calculated quickly.

    Challenge:
        - Need to obtain the 'average' CDF, and it must be very accurate.

    See [#TODO(sw 2024_11_4) put in link to pre-print] for details.
    """

    def __init__(  # noqa: PLR0913
        self,
        env_iterable: Iterable[torch.Tensor] | Iterable[list[torch.Tensor]],
        period_len: int,
        quantile: torch.Tensor | None = None,
        input_transform: InputTransform | None = None,
        outcome_transform: OutcomeTransform | None = None,
        posterior_sampler: PosteriorSampler | None = None,
        response_distribution: type[Distribution] = Gumbel,
        quantile_accuracy: torch.Tensor | None = None,
        dtype: torch.dtype = torch.float64,
        *,
        device: torch.device | None = None,
        no_grad: bool = True,
    ) -> None:
        """Initialise the QOI estimator.

        Args:
            env_iterable: An interable that produces the env data to be used. Typically this is a DataLoader.
                Supports standard env data, and weighted env data (e.g Importance sampled).

                - Standards env data: Expects tensor of shape (batch_size, d)
                - Weighted data: Expectes list of tensors.

                    - The first: the env data of shape (batch_size,d).
                    - The second item: the weight, of shape (batch_size,).

            period_len: The number of env samples in the period of interest.
            quantile: Shape (,). The quantile of the ERD to be estimate (the QoI). Should be within range (0,1).
              default .5.
            input_transform: transforms that should be applied to the env_samples before being passed to the model
            outcome_transform: transforms that should be applied to the output of the model before they are used
            posterior_sampler: The sampler to use to draw samples from the posterior of the GP.

                - ``n_posterior_samples``: is set in the PosteriorSampler
                - NOTE: if env_iterable contains batches, a batch compatible sampler such as
                  ``NormalIndependentSampler`` or "ut_sampler" should be selected.

            response_distribution: The distribution which models the stochastic response of the simulation at a single
              input point.
            quantile_accuracy: shape (,). Default value .01. Internally, optimisation is used to find the ``quantile``.
              The optimiser is allowed to terminate once in the region  ``quantile +/- quantile_accuracy``. The greater
              the accuracy required, the longer the optimisation will take. Typically other sources of uncertainty
              produce far greater uncertainty.
            dtype: The dtype used for the distribution. This has implications for the numerical accuracy possible.
            device: The device QoI should be performed on.
            no_grad: If gradient requires tracking through this QOI calculation.
        """
        self.quantile: torch.Tensor = quantile or torch.tensor(0.5, dtype=dtype)
        self.env_iterable: Iterable[torch.Tensor] | Iterable[list[torch.Tensor]] = env_iterable
        self.period_len: int = period_len
        self.input_transform: InputTransform | None = input_transform
        self.outcome_transform: OutcomeTransform | None = outcome_transform
        self.posterior_sampler: PosteriorSampler = posterior_sampler or UTSampler()
        self.response_distribution: type[Distribution] = response_distribution
        self.quantile_accuracy: torch.Tensor = quantile_accuracy or torch.tensor(0.01, dtype=dtype)
        self.dtype = dtype
        self.device: torch.device = device or torch.device("cpu")
        if self.device != torch.device("cpu"):
            msg = "GPU calcualtion are not yet supported, torch.device('cpu') must be used."
            raise NotImplementedError(msg)
        self.no_grad: bool = no_grad
        if not self.no_grad:
            msg = "Gradient tracking is not yet supported, please set `no_grad=True`."
            raise NotImplementedError(msg)

    def __call__(self, model: Model) -> torch.Tensor:
        """Estimate the QOI using the given GP model.

        Args:
            model: The GP model to use for the QOI estimation. It should have output dimension 2 which represents the
                location and scale of a Gumbel distribution.

        Returns:
            torch.Tensor: A tensor, with shape (n_posterior_samples,), of the estimated QoI for each of the
            functions sampled from the GP using the posterior sampler.
        """
        # resulting shape is (n_posterior_samples, n_env_samples, n_targets) and _posterior_samples, n_env_samples)
        params, weigths = self._parameter_estimates(model)

        if params.dtype != self.dtype or weigths.dtype != self.dtype:
            msg = (
                f"The model produced parameters of dtype {params.dtype} and weigths of dtype {weigths.dtype}."
                f" This does not match the specified dtype {self.dtype}."
                " Parameters and weights will be converted to this dtype to create a suitable distribution."
                " NOTE: underlying parameters are still only accurate to float32 precision, this may effect overall"
                " accuracy. Recommended make float64 preicitons with the model."
            )
            warnings.warn(msg, stacklevel=8)

            params, weigths = params.type(self.dtype), weigths.type(self.dtype)

        # Create the the marginal distribution for the parameters. The categorical distribution handles weights scaling
        # resulting batch shape (n_posterior_samples)
        if self.response_distribution is Gumbel:
            assert issubclass(self.response_distribution, Gumbel)  # help mypy determine the type
            loc = params[..., 0]
            scale = params[..., 1].clamp(min=1e-6)  # Ensure scale is positive
            component_dist = self.response_distribution(loc, scale)
        else:
            # To support other distribution need to to apply parameter constraints (list on the distribution), to the
            # the input params. Additionally need to know the order (name) of params in params.
            raise NotImplementedError("Only Gumbel distributions are currently supported.")

        # ApproximateMixture is appropriate becuase it produces conservative estimate. The optimisation method
        # later check that the dist has suitable resolution for the required accuracy.
        dist = ApproximateMixture(
            mixture_distribution=Categorical(weigths),
            component_distribution=component_dist,
        )

        # determine the timestep quantile to be found, and the acceptable margin of error
        _qtimestep = q_to_qtimestep(self.quantile.item(), self.period_len)
        max_qtimestep_error = acceptable_timestep_error(self.quantile.item(), self.period_len)

        # calculated the icdf values
        qtimestep = torch.tensor(_qtimestep, dtype=torch.float64)
        opt_bounds = icdf_value_bounds(dist, qtimestep)
        qois = icdf.icdf(dist, qtimestep, max_qtimestep_error, opt_bounds)
        # TODO(sw 2024-12-16): Gradient should be reattached here.
        return qois

    def _parameter_estimates(self, model: Model) -> tuple[torch.Tensor, torch.Tensor]:
        """Estimate the parameters of the individual distributions the make up the marginal distribution.

        Estimate the parameters of the component distribtuions (individual distribution that will later be marginalised)
        , and the associated mixure weights (importance weight of each of the samples in the marginalisation).

        Args:
            model: The GP model to use for the QOI estimation. The number of output distributions should match that
            required by ``response_distribution``.

        Returns:
            Tuple of tensors where:
                - Item 1: The parameters of the component distributions.
                  Shape (n_posterior_samples, n_env_samples, n_targets)
                - Item 2: The probability (importance weight) to associated with each of the component distributions.
                  Shape (n_posterior_samples, n_env_samples). The same weights are used across all posterior samples.
                  These weights are expected to be non-negative.
        """
        with ExitStack() as stack:
            # Adding context to the stack according to the configuration
            if self.no_grad:
                stack.enter_context(torch.no_grad())
            # Unclear the impact of this sett, but accoring the this shouldn't have downside. https://arxiv.org/abs/1803.06058
            stack.enter_context(gpytorch.settings.fast_pred_var())

            # make sure transforms are on the right device.
            if self.input_transform:
                assert isinstance(self.input_transform, Module)
                self.input_transform = self.input_transform.to(self.device)
            if self.outcome_transform:
                self.outcome_transform = self.outcome_transform.to(self.device)

            batched_posterior_samples_list: list[torch.Tensor] = []
            importance_weights_list: list[torch.Tensor] = []

            """NOTE: We can't easily batch the full calculation, so we just batch the posterior prediction. This can
            have runtime benefits when making the posterior predictions, and means users can provide env data in a
            format similar to other QoI methods. We then store the full dataset of importance weights and prediction,
            but this is not anticipated to be a problem, as this method does not need to use a large amount of
            ``env_samples``. Will revist this if it proves to be an issue."""
            # TODO(sw 2024-12-16): If getting too large, batch on posterior samples (make the sampler a generator?)
            for env_batch in self.env_iterable:
                # If items are lists, this is the weights are being provided (e.g Importance Sampling case).
                if isinstance(env_batch, list):
                    # The shapes should be (batch_size, d) and (batch_size,) respectively.
                    env_samples, importance_weights = env_batch
                    importance_weights_list.append(importance_weights)
                else:
                    env_samples = env_batch

                # Making sure the env_batch is on the right device
                env_samples = env_samples.to(self.device)

                if self.input_transform:
                    env_samples = self.input_transform.transform(env_samples)

                posterior = model.posterior(
                    env_samples,
                    posterior_transform=self.outcome_transform.untransform_posterior
                    if self.outcome_transform
                    else None,
                )
                # Shape is now: (n_posterior_samples, batch_size, n_targets)
                posterior_samples_batch = self.posterior_sampler(posterior)
                batched_posterior_samples_list.append(posterior_samples_batch)

            # Combines posterior samples and importance weights from all the batches
            # This is now of shape (n_posterior_samples, n_env_samples, n_targets)
            posterior_samples = torch.cat(batched_posterior_samples_list, dim=-2)

            # use importance weights if provided, otherwise treat all as equally weighted
            if importance_weights_list:
                importance_weights = torch.cat(importance_weights_list, dim=0).to(self.device)
                if importance_weights.min() < 0:
                    raise ValueError("Importance weights must be non-negative.")
            else:
                importance_weights = torch.ones(1, device=self.device, dtype=posterior_samples.dtype)

            # Provide a weight for each posterior distributionl.
            importance_weights = importance_weights.expand(posterior_samples.shape[:-1])

            return posterior_samples, importance_weights


def q_to_qtimestep(q: float, period_len: int) -> float:
    """Convert a long term quantile to an equivalent single timestep quantile.

    Args:
        q: long term quantile being estimated.
        period_len: the number of timesteps that make up the period of interest.

    Returns:
        The equivalent quantile for a single timestep with error +-

    Note:
        This funciton exists because there were numerical concerns for this process. Having a function allows us to
        document them in tests. It is appropriately accuract for very large periods. Periods of 1e13 creates an error of
        less that 1e-3 in q (the full quantile estimate). See
        `test/qoi/test_margianl_cdf_extpolation/test_q_to_qtimestep_numerical_precision_period_increase` for details.
    """
    return q ** (1 / period_len)


def acceptable_timestep_error(q: float, period_len: int, atol: float = 0.01) -> float:
    """Maximum possible single timestep error while still producing required accuracy in period estimate.

    We often make estimates for a single timestep, and scale them to a period (many timesteps). e.g.
    `q = q_timestep^period_len`. Errors in the single timestep estimates compound. This function returns the largest
    error possible to a single timestep to still be within the required accuracy in the period estimate.

    Args:
        q: long term quantile being estimated.
        period_len: the number of timesteps that make up the period of interest.
        atol: the maximum absolute error acceptable in q

    Returns:
        The maximum absolute error acceptable in the q_timestep estimate.
    """
    # NOTE: We only look at `q + atol` because this will produce a larger error than `q - atol`
    # (because of the exponential relationship)
    acceptable_error = q_to_qtimestep(q + atol, period_len) - q_to_qtimestep(q, period_len)

    FLOAT64_RESOLUTION = 1e-15  # noqa: N806
    if acceptable_error < (FLOAT64_RESOLUTION * 10):
        msg = (
            f"The acceptable error is {acceptable_error}. This is approaching or below the limits of floating point 64"
            " precision (1e-15)."
        )
        raise ValueError(msg)
    return acceptable_error



import pytest
import torch

from axtreme.utils.numerical_precision import maximal_representation_error_between_0_1


@pytest.fixture(params=[torch.float32, torch.float64])
def dtype(request: pytest.FixtureRequest) -> torch.dtype:
    return request.param


def test_maximal_representation_error_between_0_1_smaller_error_cause_no_rounding(dtype: torch.dtype):
    # representation should be insenstive to change of less than maximal representaion error
    x = torch.tensor(0.6, dtype=dtype)

    assert x + maximal_representation_error_between_0_1(x.dtype) * 0.9 == x
    assert x + maximal_representation_error_between_0_1(x.dtype) * -0.9 == x


def test_maximal_representation_error_between_0_1_larger_error_cause_rounding(dtype: torch.dtype):
    """Larger eerors are expected to cause rounding to the nex smallest representable number."""
    x = torch.tensor(0.6, dtype=dtype)

    # eps is the smallest step from [1,2). As we are in the range [.5,1), this shrink by half
    # This is the same as 2^-(mantisa + 1) as detailed in maximal_representation_error_between_0_1
    smallest_rep = torch.finfo(x.dtype).eps / 2

    # changing by maximal_representation_error will cause rounding
    assert x + maximal_representation_error_between_0_1(x.dtype) * 1.1 == x + smallest_rep
    assert x + maximal_representation_error_between_0_1(x.dtype) * -1.1 == x - smallest_rep


def test_maximal_representation_error_between_0_1_exact_error(dtype: torch.dtype):
    """The rounding behaviour when the error is exactly the maximal representation error.

    Behaviour here is a bit surpising. It doesn't round up, instead round to the binary number that is even in binary
    (last mantisa is 1).
    """

    # we pick a number that can be represented perfectly in binary. We then know the final binary digit is 0.
    x = torch.tensor(0.75, dtype=dtype)

    # Both exact error are expect to round back to the original number
    assert x + maximal_representation_error_between_0_1(x.dtype) == x
    assert x - maximal_representation_error_between_0_1(x.dtype) == x

    # eps is the smallest step from [1,2). As we are in the range [.5,1), this shrink by half
    # This is the same as 2^-(mantisa + 1) as detailed in maximal_representation_error_between_0_1
    smallest_rep = torch.finfo(x.dtype).eps / 2

    # The binary number above is odd, so this should round down to our original number
    assert x + smallest_rep - maximal_representation_error_between_0_1(x.dtype) == x

    # The binary number below is odd, so this should round up to our original number
    assert x - smallest_rep + maximal_representation_error_between_0_1(x.dtype) == x



import matplotlib.pyplot as plt
import pytest
import torch
from torch.distributions import Categorical, Distribution, Gumbel, LogNormal, MixtureSameFamily

from axtreme.distributions import ApproximateMixture
from axtreme.distributions.helpers import dist_cdf_resolution, mixture_cdf_resolution
from axtreme.distributions.mixture import icdf_value_bounds


# pyright: reportUnnecessaryTypeIgnoreComment=false
class TestApproximateMixture:
    """Testing overview: Approximate Mixture.

    Scope of testing:
        ApproximateMixture is a thin wrapper of MixtureSameFamily. We assume MixtureSameFamily test/gaurentees:

            - Produces valid CDFs and PDFs because:

                - The component distribution are gaurenteed to produce valid CDFs (from [0,1] with error outside of
                numerically supported range),
                - The mixture distribution (the weights) are gaurenteed to be in [0,1] and sum to 1.
                - The approach achieve suitable level of numerical accuracy.

        As such, the scope of testing is limited to the impact of the ApproximateMixture approximation/clipping.

    Test overview:
    - Appropriate lower/upper bound selction:
    - Running withing the bounds. Should have identical behaviour to MixtureSameFamily.
        - Single mixture.
        - batched ixture.
    - Does it perform correct clipping.
        - Large and small value expected.
        - The individual components are clipped to their specific values.
    - Weight:
        - too big and too small weights can be passed without causing issues


    Todo:
        The unit test rely on the above assumption. A developer could change the underlying implementation (so it is no
        longer ~identical to MixtueSameFamily) and break the method without any tests failing. Should implement the
        wider tests so this is not longer the case.

    """

    # TODO(sw 2024-12-14): Test more distributions
    @pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
    @pytest.mark.parametrize("dist_class", [Gumbel, LogNormal])
    def test_lower_bound_x(self, dtype: torch.dtype, dist_class: type[Distribution], *, visualise: bool = False):
        """Fnd bounds that do not throw error and produce 0."""
        dist = dist_class(torch.tensor(0, dtype=dtype), 1)  # type: ignore  # noqa: PGH003

        if visualise:
            # visualise the reverse relationship:
            finfo = torch.finfo(dtype)
            qs = torch.logspace(
                torch.log10(torch.tensor(finfo.tiny, dtype=torch.float32)),
                torch.log10(torch.tensor(0.00005, dtype=torch.float32)),
                100,
            )
            icdfs = dist.icdf(qs)

            _ = plt.axvline(finfo.eps, color="red", label="eps")
            _ = plt.scatter(qs, icdfs, label="icdf(x) -> q")
            _ = plt.ylabel("x (icdf(q)-> x)")
            _ = plt.xlabel("q ")
            _ = plt.title("ICDF\n shows the relationship in reverse")
            _ = plt.xscale("log")
            _ = plt.legend()
            plt.pause(5)

        lower_bound_x = ApproximateMixture._lower_bound_x(dist)

        assert dist.cdf(lower_bound_x) == 0

    @pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
    @pytest.mark.parametrize("dist_class", [Gumbel, LogNormal])
    def test_upper_bound_x(self, dtype: torch.dtype, dist_class: type[Distribution], *, visualise: bool = False):
        dist = dist_class(torch.tensor(0, dtype=dtype), 1)  # type: ignore  # noqa: PGH003

        if visualise:
            finfo = torch.finfo(dtype)
            max_x = dist.icdf(torch.tensor(1 - finfo.eps, dtype=dtype))
            # NOTE: these bounds may beed to be adjsuted for different dists
            x = torch.linspace(max_x - 1, max_x + 0.5, 1000, dtype=torch.float32)
            cdfs = dist.cdf(x)

            _ = plt.axhline(1 - finfo.eps, color="red", label="eps")
            _ = plt.scatter(x, cdfs, label="cdf(x)-> q")
            _ = plt.ylabel("q (cdf(x)-> q)")
            _ = plt.xlabel("x")
            _ = plt.title("CDF\nNote TransfomedDist upper bound is 1 - finfo.eps, appears to work above this point")
            _ = plt.yscale("log")
            _ = plt.legend()
            plt.pause(5)

        upper_bound_x = ApproximateMixture._upper_bound_x(dist)

        assert dist.cdf(upper_bound_x) == 1 - torch.finfo(dtype).eps

    # Nothing interesting should happen with different shapes, just checking that here.
    @pytest.mark.parametrize(
        "shape",
        [
            # Unbatch mixture distribution, with only a single component
            torch.Size([1]),
            # Batched mixture distribution, with only a single component
            torch.Size([2, 1]),
            # batch, multiple components
            torch.Size([2, 3]),
        ],
    )
    @pytest.mark.parametrize(
        "q, dtype",
        [
            (0.5, torch.float64),  # point withing the range
            (torch.finfo(torch.float64).eps, torch.float64),  # smallest value before ApproximateMixture starts clipping
            # largest value before ApproximateMixture starts clipping
            (
                1 - torch.finfo(torch.float64).eps,
                torch.float64,
            ),
        ],
    )
    def test_equivalent_to_mixturesamefamily_within_range(self, shape: torch.Size, q: float, dtype: torch.dtype):
        """ApproximateMixture and MixtureSameFamily should have identical behaviour within the component range.

        Within the range the component distribution supports, behaviour should be identical.
        NOTE: ApproximateMixture does NOT infact support the whole component range.

            - typical Component range: (finfo.tiny, 1 - finfo.eps)
            - ApproximateMixture range: (finfo.eps, 1 - finfo.eps)

        Below finfo.eps there are numerical issues with x != dist.cdf(dist.icdf(x)). Therefor ApproximateMixture clips
        this. See `docs/source/marginal_cdf_extrapolation.md` "Distribution lower bound issue" for details.

        This test checks:
            - ApproximateMixture and MixtureSameFamily have the same underling distributions
            - And the input is the in the range
            - The output should be identical
        """
        # Create the x value wich represents that quantile. The gumbel here matches the one we use as dist.
        # This has shape (1,) which can be broadcast in the CDF step to any underlying distribution
        x = Gumbel(loc=torch.zeros(1, dtype=dtype), scale=1).icdf(q)

        loc = torch.zeros(shape, dtype=dtype)
        scale = torch.ones(shape, dtype=dtype)

        dist = Gumbel(loc, scale)
        mix = Categorical(torch.ones(shape, dtype=dtype))

        approx_mixture = ApproximateMixture(mix, dist)
        mixture = MixtureSameFamily(mix, dist)

        # Run through bothe of the methods.
        actual_value = approx_mixture.cdf(x)
        expected_value = mixture.cdf(x)

        assert torch.allclose(actual_value, expected_value)

    @pytest.mark.integration  # Relies on `dist_cdf_resolution` otherwise not really an integration test.
    @pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
    def test_cdf_outside_of_component_range_single_component(self, dtype: torch.dtype):
        """Check the cdfs behaviour for x values outside of component range.

        This test uses a single underlying component distribution. This is not effected by numerical issues when taking
        weighted sum. As such finner grain tests of output can be performed. The test demonstrates the following:

            - The cdf does not fail for extremely larges of small inputs.
            - It produces the expected quantils for these results (up to the accuracy specifed by `dist_cdf_resolution`)
        """
        loc = torch.tensor([0.0], dtype=dtype)
        scale = torch.ones_like(loc, dtype=dtype)
        dist = Gumbel(loc, scale)
        mix = Categorical(torch.ones_like(loc, dtype=dtype))
        marginal = ApproximateMixture(mix, dist)

        # Find x values that have just entered the region consised outside of the component range
        finfo = torch.finfo(dist.mean.dtype)
        x_low = dist.icdf(torch.tensor(finfo.eps, dtype=dtype))
        x_high = dist.icdf(torch.tensor(1 - finfo.eps, dtype=dtype))

        max_cdf_value = marginal.cdf(x_high + x_high.abs() * 0.1)
        min_cdf_value = marginal.cdf(x_low - x_low.abs() * 0.1)

        assert min_cdf_value == pytest.approx(0, abs=dist_cdf_resolution(dist))
        assert max_cdf_value == pytest.approx(1 - finfo.eps, abs=dist_cdf_resolution(dist))

    @pytest.mark.integration  # Relies on `mixture_cdf_resolution` otherwise not really an integration test.
    @pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
    def test_cdf_outside_of_component_range_batch_mixture(self, dtype: torch.dtype):
        """Check the cdfs behaviour for x values outside of component range.

        This uses a batch of component distributions. This is effected by numerical issues when taking
        weighted sum. The test demonstrates the following:

            - The cdf does not fail for extremely larges of small inputs  (across batches)
            - It produces the expected quantiles for these results (up to the accuracy specifed by
              `mixture_cdf_resolution`). NOTE: this is a less strict bound due to our uncertainty regarding impact of
              weighted sum numerical error.
        """
        n_posterior_samples = 2
        n_env_samples_per_period = 500

        # fmt: off
        loc = torch.zeros(n_posterior_samples, n_env_samples_per_period)*0.0
        scale = torch.ones_like(loc, dtype=dtype)
        mix = Categorical(torch.ones(n_env_samples_per_period, dtype=dtype))
        # fmt: on

        comp = Gumbel(loc=loc.type(dtype), scale=scale.type(dtype))
        marginal = ApproximateMixture(mix, comp)

        # Find x values that have just entered the region consised outside of the component range
        finfo = torch.finfo(comp.mean.dtype)
        x_low = comp.icdf(torch.tensor(finfo.eps, dtype=dtype)).min()
        x_high = comp.icdf(torch.tensor(1 - finfo.eps, dtype=dtype)).max()

        # make them compatable with the batch shape of the margina
        x_low = x_low.unsqueeze(-1)
        x_high = x_high.unsqueeze(-1)

        max_cdf_value = marginal.cdf(x_high)
        min_cdf_value = marginal.cdf(x_low)

        # The lower bounds can reach 0. Their can be numerical issues that come from the weighted sum.
        # We use
        torch.testing.assert_close(
            min_cdf_value,
            torch.zeros_like(min_cdf_value),
            atol=mixture_cdf_resolution(marginal),
            rtol=0,
        )
        torch.testing.assert_close(
            max_cdf_value,
            torch.ones_like(max_cdf_value) - finfo.eps,
            atol=mixture_cdf_resolution(marginal),
            rtol=0,
        )

    def test_cdf_input_is_uniquely_clipped_for_each_comp_dist(self):
        """The input need to be clipped for the specific range each underling component supports.

        Here we check the input is clipped for one distribution and not the other. This does not cover all cases.
        """

        dtype = torch.float64
        finfo = torch.finfo(dtype)

        loc = torch.tensor([0.0, 100], dtype=dtype)
        scale = torch.ones_like(loc, dtype=dtype)
        dist = Gumbel(loc, scale)

        # find the input x to use. It should produce:
        # for the first component: cdf(x) = 1 - finfo.eps
        # for the second component: cdf(x) = .5
        x = dist.icdf(0.5)[1]

        mix = Categorical(torch.ones_like(loc))
        marginal = ApproximateMixture(mix, dist)

        expected_result_comp1 = 1 - finfo.eps
        expected_result_comp2 = 0.5
        expected_result = (expected_result_comp1 + expected_result_comp2) / 2
        actual_result = marginal.cdf(x).item()

        assert actual_result == pytest.approx(expected_result, abs=finfo.resolution)

    @pytest.mark.parametrize(
        "weights",
        [
            # total probability less than 1
            (torch.tensor([0.5, 0.3], dtype=torch.float64)),
            # total probability greater than 1
            (torch.tensor([0.5, 0.8], dtype=torch.float64)),
        ],
    )
    def test_weights_not_summing_to_one(self, weights: torch.Tensor):
        """The weights provided may be inapproapriate, check they are handled.

        Weight provided (e.g through importance sampleing) might not sum to one. Check that this distribution handles
        this and call still produce a valid distribution.

        Args:
            weights: 1d vector of weights to use in the mixture
        """
        dtype = weights.dtype

        loc = torch.zeros_like(weights)
        scale = torch.ones_like(weights, dtype=dtype)

        dist = Gumbel(loc, scale)
        mix = Categorical(probs=weights)
        marginal = ApproximateMixture(mix, dist)

        # check that appropraite CDF results are achieved
        very_large_value = torch.tensor([100_000], dtype=dtype)  # we know this is outside the bounds
        very_small_value = torch.tensor([-100_000], dtype=dtype)  # we know this is outside the bounds

        finfo = torch.finfo(dtype)
        assert marginal.cdf(very_large_value) == pytest.approx(1 - finfo.eps, abs=finfo.resolution)
        assert marginal.cdf(very_small_value) == pytest.approx(finfo.eps, abs=finfo.resolution)

    def test_cdf_input_should_have_float64_dtype(self):
        """Demonstrate how float32 input can produce different result (due to numerical issues).

        NOTE: The size of the numerical issues here are probably not large enough to cause concern
        (e.g  acceptable_timestep_error(0.5, 20 * 365 * 24, atol=0.01) = 1.13e-7), BUT we are not confident we have
        identfied the scenario causes the largest possible error.
        """
        dtype = torch.float64

        loc = torch.tensor([0.0], dtype=dtype)
        scale = torch.ones_like(loc, dtype=dtype)
        dist = Gumbel(loc, scale)
        mix = Categorical(torch.ones_like(loc))
        marginal = ApproximateMixture(mix, dist, check_dtype=False)

        x_32 = torch.tensor([0.1], dtype=torch.float32)
        x_64 = torch.tensor([0.1], dtype=torch.float64)

        q_32 = marginal.cdf(x_32)
        q_64 = marginal.cdf(x_64)

        with pytest.raises(AssertionError):
            # difference is 5.45e-10
            torch.testing.assert_close(q_32, q_64, atol=1e-10, rtol=0)


@pytest.mark.parametrize(
    "q, bounds1, bounds2",
    [
        # a simple q applied to each mixture model
        (
            torch.tensor(0.5),
            # This are known upfront becuase of the loc and scale specifiec in the test
            (Gumbel(0, 1).icdf(0.5), Gumbel(1, 1).icdf(0.5)),  # expected bound for 1s mixture model
            (Gumbel(1, 1).icdf(0.5), Gumbel(2, 1).icdf(0.5)),  # expected bound for 2nd mixture model
        ),
        # Minture model specific quantiles
        (
            torch.tensor([0.5, 0.9]),
            # This are known upfront becuase of the loc and scale specifiec in the test
            (Gumbel(0, 1).icdf(0.5), Gumbel(1, 1).icdf(0.5)),  # expected bound for 1s mixture model
            (Gumbel(1, 1).icdf(0.9), Gumbel(2, 1).icdf(0.9)),  # expected bound for 2nd mixture model
        ),
    ],
)
def test_icdf_value_bounds(q: torch.Tensor, bounds1: tuple[float, float], bounds2: tuple[float, float]):
    """Tests that specific bounds are generated for each of the underling mixture distributions.

    Check that the quantile is appropriately broacdcast.

    NOTE: Bounds are specified verbosely in areguments to make it clearer where they come from. Could simplfy this to
    just a tensor of bounds.
    """
    # Set to the require compoents
    # fmt: off
    loc = torch.tensor([[0.0, 1.,], [1.0, 2.,]], dtype=torch.float64)
    # fmt: on
    shape = torch.ones_like(loc)
    comp = Gumbel(loc=loc, scale=shape)
    mix = Categorical(torch.ones_like(loc, dtype=torch.float64))
    dist = MixtureSameFamily(mix, comp)

    # Run the test
    bounds = icdf_value_bounds(dist, q)

    # The bounds are clearly distinct, so we don't need to use high precision testing
    actual_bouds1 = bounds[:, 0]
    torch.testing.assert_close(actual_bouds1, torch.tensor(bounds1), atol=1e-5, rtol=0)
    actual_bouds2 = bounds[:, 1]
    torch.testing.assert_close(actual_bouds2, torch.tensor(bounds2), atol=1e-5, rtol=0)



import pytest
import torch
from torch.distributions import Categorical, Distribution, Gumbel, LogNormal, MixtureSameFamily

from axtreme.distributions.helpers import dist_cdf_resolution, mixture_cdf_resolution

# pyright: reportUnnecessaryTypeIgnoreComment=false


# TODO(sw 2024-12-14): Test more distributions
@pytest.mark.parametrize("dtype", [torch.float16, torch.float32])
@pytest.mark.parametrize("dist_class", [Gumbel, LogNormal])
def test_dist_cdf_resolution(dtype: torch.dtype, dist_class: type[Distribution]):
    """Bounds the numeric error of a dtype.

    A higher resolution dtype is considered to be the ground truth.
    This is a bruteforce appraoch that checks a large number of calculations.
    """
    # Set up the problem
    loc = torch.tensor([0.0], dtype=dtype)
    scale = torch.ones_like(loc, dtype=dtype)
    dist: Distribution = dist_class(loc, scale)  # type: ignore  # noqa: PGH003
    dist_ground_truth = dist_class(loc.type(torch.float64), scale.type(torch.float64))  # type: ignore  # noqa: PGH003
    finfo = torch.finfo(dtype)

    # See ApproximateMixture for why these quantiles are used as bounds
    q_lower = torch.tensor(finfo.eps, dtype=dtype)
    q_upper = torch.tensor(1 - finfo.eps, dtype=dtype)
    x_lower = dist.icdf(q_lower)
    x_upper = dist.icdf(q_upper)
    # Linspace produces nans if asked to create more possitions than the dtype supports, this heuristic avoids this
    n_points = int(min(100_000_000, 2 * finfo.max))
    x = torch.linspace(x_lower.item(), x_upper.item(), n_points, dtype=dtype)
    q = dist.cdf(x)
    q_ground_truth = dist_ground_truth.cdf(x)

    assert ((q - q_ground_truth).abs() < dist_cdf_resolution(dist)).all()


@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_mixture_dist_cdf_resolution(dtype: torch.dtype):
    """Bounds the numeric error of a dtype.

    A higher resolution dtype is considered to be the ground truth.
    This is a bruteforce appraoch that checks a large number of calculations.
    """
    # Set up the problem
    n_posterior_samples = 2
    n_env_samples_per_period = 5

    # fmt: off
    loc = torch.zeros(n_posterior_samples, n_env_samples_per_period)*0.0
    scale = torch.ones_like(loc, dtype=dtype)
    mix = Categorical(torch.ones(n_env_samples_per_period, dtype=dtype))
    # fmt: on
    comp = Gumbel(loc=loc.type(dtype), scale=scale.type(dtype))
    dist = MixtureSameFamily(mix, comp)

    mix_ground_truth = Categorical(torch.ones(n_env_samples_per_period, dtype=torch.float64))
    comp_ground_truth = Gumbel(loc=loc.type(dtype), scale=scale.type(torch.float64))
    dist_ground_truth = MixtureSameFamily(mix_ground_truth, comp_ground_truth)

    finfo = torch.finfo(dtype)

    # See ApproximateMixture for why these quantiles are used as bounds
    x_lower = comp.icdf(finfo.eps).min()
    x_upper = comp.icdf(1 - finfo.eps).max()

    # Linspace produces nans if asked to create more possitions than the dtype supports, this heuristic avoids this
    n_points = int(min(10_000_000, 2 * finfo.max))
    x = torch.linspace(x_lower.item(), x_upper.item(), n_points, dtype=dtype)
    # unsqueeze to broadcast across the batched
    x = x.unsqueeze(-1)
    q = dist.cdf(x)
    q_ground_truth = dist_ground_truth.cdf(x)

    assert ((q - q_ground_truth).abs() < mixture_cdf_resolution(dist)).all()



import pytest
import torch
from torch.distributions import Categorical, Gumbel

from axtreme.distributions import ApproximateMixture
from axtreme.distributions.utils import (
    index_batch_dist,
    index_batch_mixture_dist,
)


class TestIndexBatchDist:
    def test_index_batch_mixture_dist_standard_indexing(self):
        """Basic indexing when batch shape is 2"""
        # Set up batched dist
        loc = torch.tensor([0.0, 1.0], dtype=torch.float64)
        shape = torch.ones_like(loc)
        dist = Gumbel(loc=loc, scale=shape)

        # create underling dist
        indexed_dist = index_batch_dist(dist, (0,))

        # check they produce the same results
        x = torch.tensor([0.5], dtype=torch.float64)
        q_batch = dist.cdf(x)
        q = indexed_dist.cdf(x)

        assert q.item() == q_batch[0].item()

    def test_index_batch_mixture_dist_slicing(self):
        """Basic indexing when batch shape is [5,2]"""
        # Set up batched dist
        loc = torch.rand([5, 2, 3], dtype=torch.float64)
        shape = torch.ones_like(loc)
        dist = Gumbel(loc=loc, scale=shape)

        # create underling dist
        index = (slice(0, 2), 0)
        indexed_dist = index_batch_dist(dist, index)

        # check they produce the same results
        x = torch.tensor([0.5], dtype=torch.float64)
        q_batch = dist.cdf(x)
        q = indexed_dist.cdf(x)

        torch.testing.assert_close(q, q_batch[index], rtol=0, atol=1e-10)


class TestIndexBatchMixtureDist:
    def test_index_batch_mixture_dist_standard_indexing(self):
        """Basic indexing when batch shape is 2"""
        # Set up batched dist
        loc = torch.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]], dtype=torch.float64)
        shape = torch.ones_like(loc)
        comp = Gumbel(loc=loc, scale=shape)
        mix = Categorical(torch.ones_like(loc, dtype=torch.float64))
        dist = ApproximateMixture(mix, comp)

        # create underling dist
        indexed_dist = index_batch_mixture_dist(dist, (0,))

        # check they produce the same results
        x = torch.tensor([0.5], dtype=torch.float64)
        q_batch = dist.cdf(x)
        q = indexed_dist.cdf(x)

        assert q.item() == q_batch[0].item()

    def test_index_batch_mixture_dist_slicing(self):
        """Basic indexing when batch shape is [5,2]"""
        # Set up batched dist
        loc = torch.rand([5, 2, 3], dtype=torch.float64)
        shape = torch.ones_like(loc)
        comp = Gumbel(loc=loc, scale=shape)
        mix = Categorical(torch.ones_like(loc, dtype=torch.float64))
        dist = ApproximateMixture(mix, comp)

        # create underling dist
        index = (slice(0, 2), 0)
        indexed_dist = index_batch_mixture_dist(dist, index)

        # check they produce the same results
        x = torch.tensor([0.5], dtype=torch.float64)
        q_batch = dist.cdf(x)
        q = indexed_dist.cdf(x)

        torch.testing.assert_close(q, q_batch[index], rtol=0, atol=1e-10)

    def test_index_batch_mixture_dist_index_that_would_effect_component_distributions(self):
        """Batch same is (2,), indexing (2,1) should raise an error."""
        # Set up batched dist
        loc = torch.tensor([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]], dtype=torch.float64)
        shape = torch.ones_like(loc)
        comp = Gumbel(loc=loc, scale=shape)
        mix = Categorical(torch.ones_like(loc, dtype=torch.float64))
        dist = ApproximateMixture(mix, comp)

        # create underling dist
        with pytest.raises(IndexError, match=".*too many indice.*"):
            _ = index_batch_mixture_dist(dist, (2, 1))



import pytest
import torch
from botorch.models.deterministic import GenericDeterministicModel
from botorch.sampling.index_sampler import IndexSampler
from torch.distributions import Gumbel

from axtreme.qoi.marginal_cdf_extrapolation import MarginalCDFExtrapolation, acceptable_timestep_error, q_to_qtimestep


class TestMarginalCDFExtrapolation:
    """

    Plan:
        - __call__:
            - Unit test None
            - Itegration tests:
                - batcheed params and weights
                - dtype:
                    - float32 where safe
                    - float32 where not safe
        - _parameter_estimates:
            - Integration test:
                - batch produce same results as non batch
                - 1 and multi (2) posterior samples.
                OUT OF SCOPE:
                    - Are importance weights working properly?


    Other sub componets we potentially should test?
        - distributions with different batches get optimised/treated properly (e.g in optimisation_)
    """

    @pytest.mark.integration
    @pytest.mark.parametrize(
        "dtype, period_len",
        [
            # Short period
            (torch.float32, 3),
            (torch.float64, 3),
            # Realistic period length
            (torch.float64, 25 * 365 * 24),
        ],
    )
    def test_call_basic_example(
        self, gp_passthrough_1p: GenericDeterministicModel, dtype: torch.dtype, period_len: int
    ):
        """Runs a minimal version of the qoi using a deterministic model and a short period_len.

        Demonstrate both float32 and float64 can be used (with this short period)

        Take 2 posterior samples to comfirm the shape can be supported throughout.
        """
        # Define the inputs
        quantile = torch.tensor(0.5, dtype=dtype)
        quantile_accuracy = torch.tensor(0.5, dtype=dtype)
        # fmt: off
        env_sample= torch.tensor(
            [
                [[0], [1], [2]]
            ],
            dtype = dtype
        )
        # fmt: on

        # Run the method
        qoi_estimator_non_batch = MarginalCDFExtrapolation(
            env_iterable=env_sample,
            period_len=period_len,
            posterior_sampler=IndexSampler(torch.Size([2])),  # draw 2 posterior samples
            quantile=quantile,
            quantile_accuracy=quantile_accuracy,
            dtype=dtype,
        )
        qoi = qoi_estimator_non_batch(gp_passthrough_1p)

        # Calculated the expected value directly.
        # This relies on knowledge of the internals, specifically that the underlying distribution produce will be
        # [Gumbel(0, 1e-6), Gumbel(1, 1e-6), Gumbel(2, 1e-6)]. The first two will be clipped to qauntile q= 1-finfo.eps
        #  as per the bounds of ApproximateMixture
        dist = Gumbel(torch.tensor(2, dtype=dtype), 1e-6)
        q_timestep = (dist.cdf(qoi[0]) + (1 - torch.finfo(dtype).eps) * 2) / 3

        # check can be scaled up to the original timeframe with the desired accuracy
        assert q_timestep**period_len == pytest.approx(quantile, abs=quantile_accuracy)

    def test_call_insuffecient_numeric_precision(self, gp_passthrough_1p: GenericDeterministicModel):
        """Runs a minimal version of the qoi using a deterministic model and a short period_len.

        Demonstrate both float32 and float64 can be used (with this short period)

        Take 2 posterior samples to comfirm the shape can be supported throughout.
        """
        # Define the inputs
        dtype = torch.float32
        quantile = torch.tensor(0.5, dtype=dtype)
        quantile_accuracy = torch.tensor(0.5, dtype=dtype)
        # fmt: off
        env_sample= torch.tensor(
            [
                [[0], [1], [2]]
            ],
            dtype = dtype
        )
        # fmt: on

        # Run the method
        qoi_estimator_non_batch = MarginalCDFExtrapolation(
            env_iterable=env_sample,
            period_len=25 * 365 * 24,
            posterior_sampler=IndexSampler(torch.Size([2])),  # draw 2 posterior samples
            quantile=quantile,
            quantile_accuracy=quantile_accuracy,
            dtype=dtype,
        )

        with pytest.raises(TypeError, match="The distribution provided does not have suitable resolution"):
            _ = qoi_estimator_non_batch(gp_passthrough_1p)

    def test_parameter_estimates_env_batch_invariant(
        self, gp_passthrough_1p_sampler: IndexSampler, gp_passthrough_1p: GenericDeterministicModel
    ):
        """Checks that batched and non-batch enve data produce the saem result.

        Tests the when the data is split into smaller batches they are then combined into the same output.

        """
        # shape: (6, 1)
        # fmt: off
        env_sample_non_batch = torch.tensor(
            [
                [[0], [1], [2], [3], [4], [5]]
            ]
        )
        # fmt: on

        qoi_estimator_non_batch = MarginalCDFExtrapolation(
            env_iterable=env_sample_non_batch,
            period_len=-1,  # Not used in test
            posterior_sampler=gp_passthrough_1p_sampler,
        )
        param_non_batch, weight_non_batch = qoi_estimator_non_batch._parameter_estimates(gp_passthrough_1p)

        # shape: (2, 3, 1)
        # fmt: off
        env_sample_batch = torch.tensor(
            [
                [[0], [1], [2]],
                [ [3], [4], [5]]
            ]
        )
        # fmt: on
        qoi_estimator_batch = MarginalCDFExtrapolation(
            env_iterable=env_sample_batch,
            period_len=-1,  # Not used in test
            posterior_sampler=gp_passthrough_1p_sampler,
        )
        param_batch, weight_batch = qoi_estimator_batch._parameter_estimates(gp_passthrough_1p)

        torch.testing.assert_close(param_non_batch, param_batch)
        torch.testing.assert_close(weight_non_batch, weight_batch)


@pytest.mark.parametrize("dtype", [(torch.float32), (torch.float64)])
@pytest.mark.parametrize(
    "longterm_q, period_len",
    [
        # One hour simulation for 50 years
        (0.5, int(50 * 365.25 * 24 * 1)),
        # higher perentile
        (0.9, int(50 * 365.25 * 24 * 1)),
        # 10 minute simulation for 50 years
        (0.5, int(50 * 365.25 * 24 * 6)),
    ],
)
def test_q_to_qtimestep_numerical_precision_of_timestep_conversion(
    dtype: torch.dtype, longterm_q: float, period_len: int
):
    """Numerical stability of converting from period quantiles to timestep quantiles.

    This serves as documentation to show standard operators do not cause numerical issues when converting.
    """

    q = torch.tensor(longterm_q, dtype=dtype)
    # simple caclution
    _q_step = q_to_qtimestep(q.item(), period_len)
    q_step = torch.tensor(_q_step, dtype=dtype)

    # log based
    q_step_exp = torch.exp(torch.log(q) / period_len)

    # This is approx only for float 32, as q_to_qtimestep internally operates in float64.
    # eps is smallest representable step from 1-2, from .5 to 1 get extra bit of precision
    torch.testing.assert_close(q_step, q_step_exp)


@pytest.mark.parametrize(
    "q", [0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95]
)
def test_q_to_qtimestep_numerical_precision_period_increase(q: float):
    """Estimate the numerical error introduced through this operation.

    This calcutates the "round trip error" of going q_longterm -> q_timestep -> q_longterm, as this is easier to test.
    This error may be larger than conversion q_longterm -> q_timestep.
    By default python uses flaot64 (on most machines). This has a precision of 1e-15.

    NOTE:
        - abs = 1e-10: all tests pass
        - abs = 1e-11: japprox half the tests fail.

    By default python uses flaot64 (on most machines). This has a precision of 1e-15.
    """

    period_len = int(1e13)
    q_step = q_to_qtimestep(q, period_len)
    assert q_step**period_len == pytest.approx(q, abs=1e-3)


def test_acceptable_timestep_error_at_limits_of_precision():
    """When values reach the limits of precision check an error is thrown."""
    with pytest.raises(ValueError):
        _ = acceptable_timestep_error(0.5, int(1e6), atol=1e-10)



"""Module for the NormalIndependentSampler class.

This module contains the NormalIndependentSampler class, which is a subclass of IndependentMCSampler.
It is used to randomly generate base samples from a Normal distribution and when you want to ignore
the covariance between different x points, and sample a single X output space independently.
"""

import torch
from botorch.posteriors import GPyTorchPosterior

from axtreme.sampling.independent_sampler import IndependentMCSampler


class NormalIndependentSampler(IndependentMCSampler):
    """IndependentMCSampler that randomly generates base samples from a Normal distribution."""

    # This static method could as well be a function in this module.
    # However, it is tightly coupled with the class structure and process, so it is placed here.
    @staticmethod
    def _required_base_sample_shape(
        sample_shape: torch.Size,
        posterior: GPyTorchPosterior,
    ) -> torch.Size:
        """Determine the required shape of base_samples.

        Args:
            sample_shape: The shape (e.g 'number' but in multidim) of samples desired.
            posterior: The posterior the base samples are applicable to

        Returns:
            torch.Size of (*sample_shape, t) where:
                - t is the number of targets for that posterior
        """
        explicit_shape = NormalIndependentSampler._explicit_posterior_shape(posterior)
        target_dimensionality = explicit_shape[-1]

        return torch.Size([*sample_shape, target_dimensionality])

    def _construct_base_samples(self, posterior: GPyTorchPosterior) -> None:
        """Build the base samples that should be used for sampling the GP output produced by a single input (x).

        `IndependentMCSampler`s apply the same base samples to every x in the posterior (independantly). The base
        samples to be used are define in this method.

        Args:
            posterior: the posterior to construct the base samples for.

        Return:
            None. Stores base_samples of shape (*b,n,t) in registered buffer.
        """
        required_base_sample_shape = self._required_base_sample_shape(self.sample_shape, posterior)
        if self.base_samples is None or self.base_samples.shape != required_base_sample_shape:
            # TODO(sw): Update this if future to use a generator?
            with torch.random.fork_rng():
                _ = torch.manual_seed(self.seed)
                base_samples = torch.randn(required_base_sample_shape)

            # Consforms to the nn.Module pattern,  means this will automatically be transfered to the right device.
            self.register_buffer("base_samples", base_samples)

        if self.base_samples.device != posterior.device:
            _ = self.to(device=posterior.device)  # pragma: nocover
        if self.base_samples.dtype != posterior.dtype:
            _ = self.to(dtype=posterior.dtype)



"""Define the simulator interface downsteam axtreme components expect."""

from typing import Protocol, runtime_checkable

import numpy as np


@runtime_checkable
class Simulator(Protocol):
    """A protocol for (simulation) models."""

    def __call__(
        self, x: np.ndarray[tuple[int, int], np.dtype[np.float64]], n_simulations_per_point: int = 1
    ) -> np.ndarray[tuple[int, int, int], np.dtype[np.float64]]:
        """Evaluate the model at given points.

        Args:
            x: An array of shape (n_points, n_input_dims) of points at which to evaluate the model.
            n_simulations_per_point: The number of simulations to run at each point. Expected to have a default value.

        Returns:
            An array of shape (n_points, n_simulations_per_point, n_output_dims) of the model evaluated at the input
            points.
        """
        ...



import re
from pathlib import Path
from unittest.mock import patch

import matplotlib.pyplot as plt
import pytest
import torch
from torch.distributions import Categorical, Distribution, Gumbel, Normal

from axtreme.distributions.icdf import icdf, icdf_1d
from axtreme.distributions.mixture import ApproximateMixture
from axtreme.qoi.marginal_cdf_extrapolation import acceptable_timestep_error, q_to_qtimestep

# We need to work to high prcision to avoid numerical issues.
# TODO(sw 2024-12-11): This should eventually be moved to the specific tests as their precision requirements are known
# cdfs if float32 can give very different results to float64 in the extreme. as we often check by converting back
# through the cdf. Do not assume because because nothing fails when running float32 that there are no issues.
torch.set_default_dtype(torch.float64)


class TestICDF:
    """These the orchestration of icdf_1d. We want to test:
    - The batch (which gets flattened) ends up with results in the right place
    - the quantiles are expanded correctly
    - the bounds are expanded correctly

    TODO(sw 2024-12-12): These tests feel a little bit hacky, is there a nice way to clean them up?
    """

    def test_expanded_inputs_produce_expected_results(self):
        """Check output locations are correct when all inputs are provided in the expanded shape.

        In this test we patch the icdf_1f that is called internally, as we are only testing the orchestrartion of this
        function (e.g input and outputs all go to the right place). The method for doing this is a bit hacky.
        """

        def mock_cdf(
            dist: Distribution,
            quantile: float,
            max_acceptable_error: float,  # noqa: ARG001
            bounds: tuple[float, float],
        ) -> torch.Tensor:
            """Encodes the inputs into a single tensor which can be tracked though the function.

            This is a bit hack because we can't make a seperate element for each input becuase of the reshaping that
            occurs internally. Instead we use the different digits of a number to represent the different inputs.
            This is obviously fragile.


            Return:
                A tensor = loc + quantile + bounds[0] * 0.01 + bounds[1] * 0.001
            """
            result = dist.loc + quantile + bounds[0] * 0.01 + bounds[1] * 0.001  # type: ignore  # noqa: PGH003
            return torch.tensor([result])

        loc = torch.tensor([[1.0, 2, 3], [4, 5, 6]])
        scale = torch.ones_like(loc)
        # Batch shape(2,3)
        dist = Gumbel(loc, scale)

        # shape: 2,3
        quantile = loc * 0.1
        max_acceptable_error = 1e-5
        # shape (2,2, 3)
        bounds = torch.tensor([[[1, 2, 3], [4, 5, 6]], [[1, 2, 3], [4, 5, 6]]])

        with patch("axtreme.distributions.icdf.icdf_1d", new=mock_cdf):
            result = icdf(dist, quantile, max_acceptable_error, bounds)

        assert result.shape == (2, 3)

        # same asL loc + quantile + bounds[0] * 0.01 + bounds[1] * 0.001
        expected_results = torch.tensor([[1.1110, 2.2220, 3.3330], [4.4440, 5.5550, 6.6660]])
        torch.testing.assert_close(result, expected_results)

    @pytest.mark.parametrize(
        "quantile, expected_results",
        [
            (torch.tensor(0.9), torch.tensor([[1.9110, 2.9220, 3.9330], [4.9440, 5.9550, 6.9660]])),
            (torch.tensor([[0.8], [0.9]]), torch.tensor([[1.8110, 2.8220, 3.8330], [4.9440, 5.9550, 6.9660]])),
        ],
    )
    def test_q_shapes(self, quantile: torch.Tensor, expected_results: torch.Tensor):
        """Duplicate of the above, but with different shapes of q. This is to test that the q is expanded correctly."""

        def mock_cdf(
            dist: Distribution,
            quantile: float,
            max_acceptable_error: float,  # noqa: ARG001
            bounds: tuple[float, float],
        ) -> torch.Tensor:
            """Encodes the inputs into a single tensor which can be tracked though the function.

            This is a bit hack because we can't make a seperate element for each input becuase of the reshaping that
            occurs internally. Instead we use the different digits of a number to represent the different inputs.
            This is obviously fragile.


            Return:
                A tensor = loc + quantile + bounds[0] * 0.01 + bounds[1] * 0.001
            """
            result = dist.loc + quantile + bounds[0] * 0.01 + bounds[1] * 0.001  # type: ignore  # noqa: PGH003
            return torch.tensor([result])

        loc = torch.tensor([[1.0, 2, 3], [4, 5, 6]])
        scale = torch.ones_like(loc)
        # Batch shape(2,3)
        dist = Gumbel(loc, scale)

        max_acceptable_error = 1e-5
        # shape (2,2, 3)
        bounds = torch.tensor([[[1, 2, 3], [4, 5, 6]], [[1, 2, 3], [4, 5, 6]]])

        with patch("axtreme.distributions.icdf.icdf_1d", new=mock_cdf):
            result = icdf(dist, quantile, max_acceptable_error, bounds)

        # same asL loc + quantile + bounds[0] * 0.01 + bounds[1] * 0.001
        torch.testing.assert_close(result, expected_results)

    @pytest.mark.parametrize(
        "bounds, expected_results",
        [
            (torch.tensor([8, 9]).reshape(2, 1, 1), torch.tensor([[1.1890, 2.2890, 3.3890], [4.4890, 5.5890, 6.6890]])),
            (
                torch.tensor([[6, 8], [7, 9]]).reshape(2, 2, 1),
                torch.tensor([[1.1670, 2.2670, 3.3670], [4.4890, 5.5890, 6.6890]]),
            ),
        ],
    )
    def test_bounds_shapes(self, bounds: torch.Tensor, expected_results: torch.Tensor):
        """Duplicate of the above, but with different shapes of bounds. This is to test that the bounds are expanded
        correctly.
        """

        def mock_cdf(
            dist: Distribution,
            quantile: float,
            max_acceptable_error: float,  # noqa: ARG001
            bounds: tuple[float, float],
        ) -> torch.Tensor:
            """Encodes the inputs into a single tensor which can be tracked though the function.

            This is a bit hack because we can't make a seperate element for each input becuase of the reshaping that
            occurs internally. Instead we use the different digits of a number to represent the different inputs.
            This is obviously fragile.


            Return:
                A tensor = loc + quantile + bounds[0] * 0.01 + bounds[1] * 0.001
            """
            result = dist.loc + quantile + bounds[0] * 0.01 + bounds[1] * 0.001  # type: ignore  # noqa: PGH003
            return torch.tensor([result])

        loc = torch.tensor([[1.0, 2, 3], [4, 5, 6]])
        scale = torch.ones_like(loc)
        # Batch shape(2,3)
        dist = Gumbel(loc, scale)
        quantile = loc * 0.1
        max_acceptable_error = 1e-5

        with patch("axtreme.distributions.icdf.icdf_1d", new=mock_cdf):
            result = icdf(dist, quantile, max_acceptable_error, bounds)

        # same asL loc + quantile + bounds[0] * 0.01 + bounds[1] * 0.001
        torch.testing.assert_close(result, expected_results)


class TestICDF1D:
    """Testing rational: icdf_1d
    This uses an existing optimiser. We are not testing if the optimiser itselfs work, we are testing if we have
    configured it correctly, and it is suitable for our problem. We only test it on variant that we think are relative
    to our problem:


    Assumptions:

        - It is a valid CDF (Continous, monotonically increasing, bounded by [0,1]

    Cases relative to our problem:

        - Case 1: Optimisation in the extreme/tail region of distrbutions of standard distributions.
        - Case 2: Optimisation of a Mixture distribution (representing the marginal CDF)

            - Unit test: Function/line properties we could extect.

                1) Basic/smooth function [Covered in Case 1]
                2) Small discrete steps
                3) Location ands scale produced by a demo problem
                4) Incorrect bounds
                5) Dtype with insuffecient precision
                7)


            - Integration testing: We could do extensive testing by making combination of:

                - Distribution represetent the relationship between location and scale (e.g big loc, small scale)
                - make this function noise
                - See if the optimisation result reproduces the brute force

                NOTE: This is currently out of scope.


    Further tests requried and open questions:
        - Create test cases for:
            - scipy hitting x accuracy limit
            - optimisation terminating with insuffecient accuracy
            - optimisation clipping
        - what is the impact of using float32.
            - Do the number change significantly in test_icdf_1d_approximate_mixture_demo_problem_params. Compare this
              with
            the example in "demo/qoi_testing_extrapolation.py"
            - step size in q What happens if the optimise cant get to required precision because steps are too big
            - step size in x. What happens if the optimiser can get to the required accuracy because steps are too big
            - can there be a wide range of y values that return the same q. What result should we return in this case.
            - errors compound (e.g representation error +* max_acceptable_error), we don't consider how this impacts the
            total error
    """

    @pytest.mark.parametrize(
        "dist, q, maximum_timestep_error, bounds",
        [
            # Basic test - not realistic for our usecase
            (Gumbel(0, 1), 0.5, 1e-5, (-2, 5)),
            # Case 1: Optimisation in tail region. q - .5 at 20 years of 1 hour timesteps
            (
                Gumbel(0, 1),
                0.9999960436883025,
                acceptable_timestep_error(0.5, 20 * 365 * 24, atol=0.01),
                (-3, 30),
            ),
            (
                Normal(0, 1),
                0.9999960436883025,
                acceptable_timestep_error(0.5, 20 * 365 * 24, atol=0.01),
                (-30, 30),
            ),
        ],
    )
    def test_icdf_1d_basic_distribution(
        self, dist: Distribution, q: float, maximum_timestep_error: float, bounds: tuple[float, float]
    ):
        """Corresponds to "case 1" in the "testing rational".

        NOTE: Bounds are picked to be very wide without failing because they are outside the distribution numerical
        bounds.
        """
        actual_x = icdf_1d(dist, q, maximum_timestep_error, bounds)
        actual_q = dist.cdf(actual_x)

        assert actual_q.item() == pytest.approx(q, abs=maximum_timestep_error)

    @pytest.mark.integration
    def test_icdf_1d_approximate_mixture_step_function(self, *, visualise: bool = False):
        """Tests persormance on a challening Approximate Mixture distributiion.

        This is case 2.2 in test rational. This is a challening distribution that is effectly a step function.
        the desired quantil is effectively on the "corner" of one of the steps.
        """
        q_timestep = 0.9999960436883025  # equivalent to q_to_qtimestep(0.5, 20 * 365 * 24)
        maximum_timestep_error = acceptable_timestep_error(0.5, 20 * 365 * 24, atol=0.01)

        loc = torch.tensor([0, 1, 2, 3, 4, 10], dtype=torch.float64)
        scale = torch.ones_like(loc) * 0.001
        comp = Gumbel(loc, scale)
        mix = Categorical(torch.ones_like(loc, dtype=torch.float64))
        dist = ApproximateMixture(mix, comp)

        if visualise:
            x_domain = torch.linspace(0, 12, 100, dtype=torch.float64)
            y = dist.cdf(x_domain)
            _ = plt.plot(x_domain, y)
            _ = plt.xlabel("x")
            _ = plt.ylabel("CDF")
            plt.pause(5)

        # this brute force is accurate to 1e-10 q
        x_brute_force = torch.tensor(10.010648506485065, dtype=torch.float64)
        actual_x = icdf_1d(dist, q_timestep, maximum_timestep_error, bounds=(0, 20))
        # The actual value has some allowable error which means it does not have to be the brute force result exactly.
        # The below threshold is choosen arbitraily, the extent of the approximation is tested in the following assert/
        torch.testing.assert_close(actual_x, x_brute_force, rtol=0, atol=1e-6)

        # check the error
        q_error_actual = q_timestep - dist.cdf(actual_x)
        assert q_error_actual.abs().item() < maximum_timestep_error

    @pytest.mark.integration
    def test_icdf_1d_approximate_mixture_demo_problem_params(self):
        """Tests persormance on a challening Approximate Mixture distributiion.

        This is case 2.3 in test rational.

        NOTE: This is a good usecase for exploring the impacts of float32 as it appear to change the result
        significantly.
        """
        # define the required accuracy
        q_timestep = 0.9999960436883025  # equivalent to q_to_qtimestep(0.5, 20 * 365 * 24)
        maximum_timestep_error = acceptable_timestep_error(0.5, 20 * 365 * 24, atol=0.01)

        # create the distribution
        data_dir = Path(__file__).parent / "data" / "test_icdf"
        loc = torch.load(data_dir / "demo_loc.pt", weights_only=True).type(torch.float64)
        scale = torch.load(data_dir / "demo_scale.pt", weights_only=True).type(torch.float64)
        comp = Gumbel(loc, scale)
        mix = Categorical(torch.ones_like(loc, dtype=torch.float64))
        dist = ApproximateMixture(mix, comp)

        actual_x = icdf_1d(dist, q_timestep, maximum_timestep_error, bounds=(0, 100))

        x_brute_force = torch.tensor(61.025285285285285, dtype=torch.float64)
        # The actual value has some allowable error which means it does not have to be the brute force result exactly.
        # The below threshold is choosen arbitraily, the extent of the approximation is tested in the following assert/
        torch.testing.assert_close(actual_x, x_brute_force, rtol=0, atol=1e-5)

        # check the error
        q_error_actual = q_timestep - dist.cdf(actual_x)
        assert q_error_actual.abs().item() < maximum_timestep_error

    def test_icdf_1d_respects_maximal_error(self):
        # make a required error
        period_len = 20 * 365 * 24
        quantile = 0.5
        quantile_accuracy = 0.01
        q_timestep = q_to_qtimestep(quantile, period_len)
        maximum_timestep_error = acceptable_timestep_error(0.5, period_len, atol=quantile_accuracy)

        dist = Gumbel(torch.tensor([0], dtype=torch.float64), 1)

        actual_x = icdf_1d(dist, q_timestep, maximum_timestep_error, bounds=(0, 13))

        actual_q = dist.cdf(actual_x)

        assert actual_q**period_len == pytest.approx(quantile, abs=quantile_accuracy)

    def test_icdf_1d_incorret_bounds(self):
        """Optimiation should fail and return useful information when the bounds are incorrect."""
        dist = Gumbel(0, 1)
        q = 0.9999960436883025  # equaivalent to q_to_qtimestep(0.5, 20 * 365 * 24)
        maximum_timestep_error = acceptable_timestep_error(0.5, 20 * 365 * 24, atol=0.01)
        bounds = (20, 30)

        with pytest.raises(ValueError, match=re.escape("f(a) and f(b) must have different signs")):
            _ = icdf_1d(dist, q, maximum_timestep_error, bounds)

    def test_icdf_1d_incorrect_dtype(self):
        """Test the dtype being used is approapriate for the problem.

        The dtype used by torch can have a significant effect on the calculations. This is because the result is
        required to be accurate to many decimal place (the quantile is taken to the power many times). Specifically
        torch.float32 is typically not enough.

        General considertion for the optimisation:
            - Outputs don't have sufficient precision. (the dist is of the wrong dtype)
            - Input being optimised don't have suffecient precision

            These can cause it to fail due to not reaching the appropriate precision. They should be highlighted.

        Specific to scipy:
            - inputs are controlled by scipy and are always float. This is not a concern.
            - Outputs. Unclear what happen when unsuitable precision is used. TODO(sw)
        """

        loc = torch.tensor([0], dtype=torch.float32)
        scale = torch.ones_like(loc)
        dist = Gumbel(loc, scale)

        q = 0.9999960436883025  # equivalent to q_to_qtimestep(0.5, 20 * 365 * 24)

        with pytest.raises(TypeError, match="distribution provided does not have suitable resolution"):
            _ = icdf_1d(dist, q, 1e-10, (3, 13))



import numpy as np
from ax.core.observation import Observation, ObservationData, ObservationFeatures

from axtreme.utils.modelbridge_utils import observations_to_arrays


def test_observations_to_arrays():
    x = np.array([[1, 2], [3, 4], [5, 6]])
    y = np.array([[10], [20], [30]])
    of = [ObservationFeatures(parameters={"x1": point[0], "x2": point[1]}) for point in x]
    od = [ObservationData(metric_names=["y"], means=point, covariance=np.eye(1)) for point in y]
    observations = [Observation(features=f, data=d) for f, d in zip(of, od, strict=False)]
    features, f, cov = observations_to_arrays(param_names=["x1", "x2"], outcomes=["y"], observations=observations)
    assert np.array_equal(features, x)
    assert np.array_equal(f, y)
    # There is only one output per x point. This will have perfect covariance with itself.
    assert np.array_equal(cov, np.ones([3, 1, 1]))



"""Plotting helper tailored for analysingQoIJobResults."""

import copy
from collections.abc import Callable

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from botorch.models import PosteriorMeanModel, SingleTaskGP
from botorch.sampling.index_sampler import IndexSampler
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from pandas.core.groupby.generic import DataFrameGroupBy

from axtreme.qoi.gp_bruteforce import GPBruteForce


def plot_distribution(
    df: pd.DataFrame,
    ax: Axes,
    n_hists: int = 3,
    col_name: str = "samples",
    brute_force: float | None = None,
) -> None:
    """Helperfor plotting histograms (on the given ax) of dataframe cells containing lists.

    Designed for use with the 'samples' column of a QoiJobResults dataframe.

    Args:
        df: A dataframe.
        ax: The axis to plot on.
        n_hists: The number of cells of column `col_name` to plot.
        col_name: The column of the df containing lists.
        brute_force: Represents the true value (e.g mean). Plots a vertical line if provided.
    """
    samples_list = df.loc[:, col_name].to_numpy()
    samples_list = samples_list[:n_hists]

    for samples_i in samples_list:
        _ = ax.hist(samples_i, density=True, alpha=0.3, bins=len(samples_i) // 5 + 1)
    _ = ax.set_xlabel("QOI value")
    _ = ax.set_title("QOi estimator distibutions")

    if brute_force:
        _ = ax.axvline(brute_force, c="black", label=f"Brute force ({brute_force:.2f})")

    _ = ax.legend()


def plot_col_histogram(df: pd.DataFrame, ax: Axes, col_name: str = "mean", brute_force: float | None = None) -> None:
    """Helper which creates a histogram (on the given ax) based on a column of the df.

    Designed for use with the 'mean' or 'var' column of a QoiJobResults dataframe.

    Args:
        df: A dataframe.
        ax: The axis to plot on.
        col_name: The column of the df containing lists.
        brute_force: Represents the true value (e.g mean). Plots a vertical line if provided.
    """
    values = df.loc[:, col_name].to_numpy()
    _ = ax.hist(values, label=f"{col_name} of qoi runs", density=True, bins=30)
    title_str = (
        f"{col_name} of each qoi run\n"
        f"mean of dist {values.mean():.3f}."
        f" std of dist {values.std():.3f},"
        # Protect against divide by error error
        f"C.o.V {values.std()/ values.mean() if values.mean() > 1e-2 else np.nan:.3f}"  # noqa: PLR2004
    )
    _ = ax.set_title(title_str)
    _ = ax.set_ylabel("density")
    _ = ax.set_xlabel("QOI value")

    if brute_force:
        _ = ax.axvline(brute_force, c="black", label=f"Brute force ({brute_force:.2f})")

    _ = ax.legend()


def plot_groups(
    # Error if try to parmaterise this as a generic
    df_grouped: DataFrameGroupBy,  ## pyright: ignore[reportUnknownParameterType,reportMissingTypeArgument]
    plotting_funcs: list[Callable[[pd.DataFrame, Axes], None]],
) -> Figure:
    """Takes a grouped dataframe, and generates a row of plots for each group, using plotting_funcs.

    Args:
        df_grouped: The groupby object to plot
        plotting_funcs: list of plots to be generated for each group. See plot_col_histogram for an example of a
          plotting function.
    """
    group_names = df_grouped.groups
    n_groups = len(group_names)
    n_cols = len(plotting_funcs)
    fig, axes = plt.subplots(nrows=n_groups, ncols=n_cols, figsize=(6 * n_cols, 4 * n_groups), sharex="col")

    if n_groups == 1:
        # reshape so can be handled in the same manner as 2d
        axes = axes.reshape(1, n_cols)

    for i, (_, group_data) in enumerate(df_grouped):
        row_axes = axes[i]
        for plot_func, ax in zip(plotting_funcs, row_axes, strict=True):
            plot_func(group_data, ax)

    # Add group labels to each row
    for ax, row in zip(axes[:, 0], group_names, strict=True):
        ax.annotate(
            row,
            xy=(0, 0.5),
            xytext=(-ax.yaxis.labelpad - 5, 0),
            xycoords=ax.yaxis.label,
            textcoords="offset points",
            size="large",
            ha="right",
            va="center",
        )

    fig.tight_layout()
    # tight_layout doesn't take these labels into account. Manually tweak.
    fig.subplots_adjust(left=0.15)

    return fig


def qoi_ignoring_gp_uncertainty(qoi: GPBruteForce, model: SingleTaskGP) -> torch.Tensor:
    """Helper to run a QoI with a model, ignoring uncertainty in the model (e.g using the posterior mean).

    Args:
        qoi: The QoI estimator to use
        model: The model to use

    Return:
        the estimates made by the QoI using only the posterior mean of the model.
    """
    # copy the QoI as we will need to change some internal settings
    new_qoi = copy.deepcopy(qoi)

    # make a deterministic model
    det_model = PosteriorMeanModel(model)

    # We need to change the sampler to be compatible with a deterministic model

    if not hasattr(new_qoi.posterior_sampler, "sample_shape"):
        msg = (
            "This function only works with samplers that have a sample_shape attribute."
            " e.g Like those inheriting from botorch MCSampler."
        )
        raise NotImplementedError(msg)

    sample_shape = new_qoi.posterior_sampler.sample_shape  # pyright: ignore[reportAttributeAccessIssue]
    sampler = IndexSampler(sample_shape)
    # We need to do the following work around to make IndexSampler compatible with TransformedPosterior. IndexSampler is
    # only designed to work with EnsembelPosterior, but when using outcome transforms on the model, the
    # EnsembelPosterior gets wrapped in a TransformedPosterior. This only causes issues when base samples are created
    # internally, so we create them here instead.
    # TODO(sw 2024-11-22): raise a ticket with botorch to fix this issue?
    sampler.register_buffer("base_samples", torch.zeros(sample_shape, dtype=torch.int))

    new_qoi.posterior_sampler = sampler

    return new_qoi(model=det_model)



"""Additional Metric implemenations."""

from typing import TYPE_CHECKING, Any

import pandas as pd
from ax import Data, Metric
from ax.core.base_trial import BaseTrial
from ax.core.metric import MetricFetchResult
from ax.utils.common.result import Ok

if TYPE_CHECKING:
    from axtreme.evaluation import SimulationPointResults


class LocalMetadataMetric(Metric):
    """This metric retrieves its results form the trial metadata.

    The simple example run the simultion within this function call
    (e.g. `see <https://ax.dev/tutorials/gpei_hartmann_developer.html#8.-Defining-custom-metrics>`_)
    In general, this method should only 'fetch' the results from somewhere else where they have been run.
    For example, Runner deploys simulation of remote, this connects to remote and collects result.
    This is local implementation of this patter, where results are stored on trail metadata.

    This is useful when:
    - Running a single simulation produces multiple output metrics
    (meaning the simulation doesn't need to be run as many times)
    - Want to execute the simulation when `trail.run()` is called

    Note:
        This object is coupled with LocalMetadataRunner, through SimulationPointResults

    Background:

    Flow:
    - trial.run() called, internally call the runner, and puts the resulting data into metadata
    - Later Metric.fetch_trial_data(trial) is called Therefore, Metric has access to all the "bookkeeping"
    trial information directly, the only thing that should be in metadata is run result.
    """

    # TODO(sw): Make it explictly support single arm trial
    def fetch_trial_data(  # noqa: D102
        self,
        trial: BaseTrial,
        **kwargs: Any,  # noqa: ANN401, ARG002 NOTE: kwargs is needed to match the signature of the parent class.
    ) -> MetricFetchResult:
        records = []
        for arm_name in trial.arms_by_name:
            point_result: SimulationPointResults = trial.run_metadata["simulation_result"]
            metrics_columns = point_result.metric_data(self.name)
            # The data that must be contained in the results can be found here: Data.REQUIRED_COLUMNS
            # Required keys are {"arm_name","metric_name", "mean", "sem"}
            # Additional info can be found here: ax.core.data.BaseData
            records.append(
                {
                    "arm_name": arm_name,
                    "trial_index": trial.index,
                    "metric_name": self.name,
                    **metrics_columns,
                }
            )

        return Ok(value=Data(df=pd.DataFrame.from_records(records)))



"""GPBruteForce calculates QoIs by applying the surrogate to each env sample in the period."""
# %%
# pyright: reportUnnecessaryTypeIgnoreComment=false

import warnings
from collections.abc import Iterable
from contextlib import ExitStack
from typing import TypeVar

import gpytorch
import torch
from botorch.models.model import Model
from botorch.models.transforms.input import InputTransform
from botorch.models.transforms.outcome import OutcomeTransform
from torch.distributions import Gumbel
from torch.nn.modules import Module

from axtreme.qoi.qoi_estimator import QoIEstimator
from axtreme.sampling import MeanVarPosteriorSampledEstimates, NormalIndependentSampler, PosteriorSampler

T = TypeVar("T")


class GPBruteForce(MeanVarPosteriorSampledEstimates, QoIEstimator):
    """Estimate the QoI for an extreme response distribution, using a surrogate model.

    Uses a full periods of environment samples, and passes each sample through the surrogate.

    Overview of the algorithem:

    - Take `n` periods of environment data points.
    - Use the surrogate model to estimate the likely response distribution at each point (the posterior).
    - Take `n_posterior_samples` of the posterior, each representing a guess at what the true simulator is.

    - For each posterior sample:

      - Simulate the response seen at each data point.
      - Find the largest response in each period. Each of these is a sample of that posterior's ERD.
      - Calculate the QoI based on these ERD samples.

    - Return the QoIs calculated by each posterior sample.

    Uncertainty in the results comes from three sources:

    - The envirionment samples used.
    - Uncertainty in the GP and the posterior samples used.
    - Randomness from sampling the surrogates output distribution.

    Optimisation Notes:
        GPBruteForce is not smooth w.r.t to smooth changes in the model (e.g like provided By QoILookAhead).

    Todo:
        Provide reference to the gpbruteforce.md file in docs so it renders (sw 2024-11-29).
    """

    def __init__(  # noqa: PLR0913
        self,
        env_iterable: Iterable[torch.Tensor],
        input_transform: InputTransform | None = None,
        outcome_transform: OutcomeTransform | None = None,
        posterior_sampler: PosteriorSampler | None = None,
        *,
        erd_samples_per_period: int = 1,
        shared_surrogate_base_samples: bool = False,
        device: torch.device | None = None,
        no_grad: bool = True,
        seed: int | None = None,
    ) -> None:
        """Initialise the QOI estimator.

        Args:
            env_iterable: An iterable that produces the env data to be used. Typically this is a DataLoader.

                - The iterable contains batches of shape (n_periods, batch_size, d).
                - Combining all of the batch should produce the shape (n_periods, period_len, d).
                - This is an iterable because predictions often need to be made in batches for memory reasons.
                - If your data is small, you can process it all at once by passing `[data]`, where `data` is a tensor.

            input_transform: Transforms that should be applied to the env_samples before being passed to the model.
            outcome_transform: Transforms that should be applied to the output of the model before they are used.
            posterior_sampler: The sampler to use to draw samples from the posterior of the GP.

                - `n_posterior_samples` is set in the PosteriorSampler.

                .. Note::
                    If `env_iterable` contains batches, a batch-compatible sampler, such as
                    `NormalIndependentSampler`, should be chosen.

            erd_samples_per_period: Number of ERD samples created from a single period of data. This can reduce the
                noise of sampling the response drawn from the surrogate's response distribution (at a point 'x').
            shared_surrogate_base_samples: If True, all `n_posterior_samples` will use the same base samples when
                sampling the surrogate's response output. As a result, the posterior samples are responsible for any
                difference in ERD distribution (e.g., surrogate sampling noise no longer contributes).

                - Set to False: Better shows overall uncertainty in QoI.
                - Set to True: Shows only uncertainty caused by GP uncertainty.

            device: The device that the model should be run on.
            no_grad: Whether to disable gradient tracking for this QOI calculation.
            seed: The seed to use for the random number generator. If None, no seed is set.
        """
        super().__init__()
        self.env_iterable = env_iterable
        self.input_transform: InputTransform | None = input_transform
        self.outcome_transform: OutcomeTransform | None = outcome_transform
        self.posterior_sampler: PosteriorSampler = posterior_sampler or NormalIndependentSampler(torch.Size([100]))
        self.erd_samples_per_period: int = erd_samples_per_period
        self.shared_surrogate_base_samples = shared_surrogate_base_samples
        self.device: torch.device = device or torch.device("cpu")
        if self.device != torch.device("cpu"):
            msg = "GPU calcualtion are not yet supported, torch.device('cpu') must be used."
            raise NotImplementedError(msg)
        self.no_grad: bool = no_grad
        if not self.no_grad:
            msg = "Gradient tracking is not yet supported, please set `no_grad=True`."
            raise NotImplementedError(msg)
        self.seed: int | None = seed

    def __call__(self, model: Model) -> torch.Tensor:
        """Estimate the QOI using the given GP model.

        Args:
            model: The GP model to use for the QOI estimation. It should have output dimension 2 which represents the
                location and scale of a Gumbel distribution.

        Returns:
            Tensor: 1d tensor where each entry is an estimate of the QOI as produced by function sampled from the GP.
            The function sampled from the GP (and the number of them) is determined by `posterior_sampler`.
        """
        posterior_samples_erd_samples = self.posterior_samples_erd_samples(model)

        n_erd_samples = posterior_samples_erd_samples.shape[-1]
        if n_erd_samples % 2 == 0:
            msg = (
                f"There are {n_erd_samples} erd samples for each sample frrom the GP (posterior samples)."
                " tensor.median() called on a even number of samples produces a systematic underestimate of the median"
                " (because `torch.tensor([1,2,3,4]).median() -> 2`). n_erd_samples is determined by n_periods"
                " * erd_samples_per_period (where n_periods=env_iterable.shape[-3])"
            )
            warnings.warn(msg, UserWarning, stacklevel=2)

        # TODO(ks): This should maybe be changed to include different statistic than the median (any quantile?)
        # Would need to update the warning when changing this
        return posterior_samples_erd_samples.median(dim=-1).values

    def posterior_samples_erd_samples(self, model: Model) -> torch.Tensor:
        """Returns the erd samples created by each posterior sample.

        __call__ uses these erd sample to create a QoI estimate per posterior.

        Args:
            model: The GP model to use for the QOI estimation. It should have output dimension 2 which represents the
                location and scale of a Gumbel distribution.

        Returns:
            Tensor: The erd samples obtained for each function (posterior sample) obtianed from the GP.
            Shape: (n_posterior_samples, n_periods * erd_samples_per_period)
        """
        with ExitStack() as stack:
            # Adding context to the stack according to the configuration
            if self.no_grad:
                stack.enter_context(torch.no_grad())
            if self.seed is not None:
                stack.enter_context(torch.random.fork_rng())

                # If a seed is provided, set the seed for the random number generator
                # This is done in the context manager to ensure that the seed is only set for this calculation
                _ = torch.manual_seed(self.seed)
            stack.enter_context(gpytorch.settings.fast_pred_var())

            # make sure transforms are on the right device.
            if self.input_transform:
                assert isinstance(self.input_transform, Module)
                self.input_transform = self.input_transform.to(self.device)
            if self.outcome_transform:
                self.outcome_transform = self.outcome_transform.to(self.device)

            # This will track the most extreme value seen thus far in the iterations
            extreme_value_thus_far: torch.Tensor | None = None
            for batch in self.env_iterable:
                # This is of shape: (n_periods, batch_size, d)
                env_batch = batch.to(self.device)

                # This has shape: (erd_samples_per_period, n_posterior_samples, n_periods)
                batch_extreme_responses = self._process_batch(env_batch, model)

                # If a bigger extreme value has been seen, replace the tracked value
                extreme_value_thus_far = (
                    extreme_value_thus_far.max(batch_extreme_responses)
                    if extreme_value_thus_far is not None
                    else batch_extreme_responses
                )

            if extreme_value_thus_far is None:
                msg = "env_iterable provided was empty"
                raise ValueError(msg)

            # Group the ERD samples that belong to the same posterior sample
            # Shape is now:  (n_posterior_samples, n_periods * erd_samples_per_period)
            extreme_value_thus_far = torch.cat([*extreme_value_thus_far], dim=-1)
            return extreme_value_thus_far

    @staticmethod
    def sample_surrogate(
        params: torch.Tensor, n_samples: int = 1, base_sample_broadcast_dims: list[int] | None = None
    ) -> torch.Tensor:
        """Create the surrogate model for a given set of input parameters, and sample response of the surrogate.

        Typically a GP is used to parameterise the surrogate model at a specific x. The now parameterise model can be
        run multiple times to get different realisations of the stochastic response.

        Args:
            params: (`*b`, p) tensor of parameters. The last dimesion is expected to contain the parameters required to
            instantiate a single surrogate model. All other dimensions are optional batch dimension.
            n_samples: The number of samples to draw from the surrogate model at a single x point.
            base_sample_broadcast_dims: List of indexes in (`*b`). Base samples will be shared (broadcast) across these
                dimension of `*b`. For example:

                - params.shape is (n_posterior_samples, n_periods, batch_size, n_params).

                  - `*b` = (n_posterior_samples, n_periods, batch_size)
                  - p = (n_params)

                - You would like to use the same base samples for each n_posterior_samples, so that any difference in
                  output can be attributed to the difference in the n_params, rather than due to the randomness in the
                  sample generated by the surrogate mode.

                - By setting `base_sample_broadcast_dims=[0]` the base samples used would be of shape
                  (1, n_periods, batch_size), which would achieve the desired effect.

        Returns:
            tensor of (n_samples, `*b`) representing the response of the surrogate model.

        Todo:
            The `base_sample_broadcast_dims` behaviour is challenging to describe now that is in this function rather
            than in context. Alternately base samples could be directly applied like in
            posterior.rsample_from_base_samples. We have avoid this so complexity is contained here for now.
        """
        gumbel_dists = Gumbel(
            params[..., 0],
            params[..., 1].clamp(min=0.0000001),
            validate_args=True,
        )

        b_shape = list(params.shape[:-1])
        # Change dimension so they will be broadcast if required.
        if base_sample_broadcast_dims:
            for idx in base_sample_broadcast_dims:
                b_shape[idx] = 1

        base_samples_shape = torch.Size([n_samples]) + tuple(b_shape)

        # This is a home baked version of gumbel_dists.rsample_from_base_samples
        # broadly speaking it produces the same results as gumbel_dists.rsample(n_samples)
        base_samples = torch.rand(base_samples_shape)
        gumbel_samples = gumbel_dists.icdf(base_samples)

        return gumbel_samples

    def _process_batch(self, env_batch: torch.Tensor, model: Model) -> torch.Tensor:
        """Process batches of env_data, returning the max response seen across each period.

        For each period, n_posterior_samples provide different plausable estimates of the response distribution.
        erd_samples_per_period are then drawn from each response distribution. Each of these is stored seperately so
        they can later be aggregated across batches.

        Args:
            env_batch: (n_periods, batch_size, d)
            model: The gp to be used as the surrogate model

        Return:
            (erd_samples_per_period, n_posterior_samples, n_periods)
        """
        if self.input_transform:
            env_batch = self.input_transform.transform(env_batch)

        posterior = model.posterior(
            env_batch,
            posterior_transform=self.outcome_transform.untransform_posterior if self.outcome_transform else None,
        )

        # Sampling from the posterior using the provided sampler
        # shape is now: (n_posterior_samples, n_periods, batch_size, n_targets)
        gp_samples: torch.Tensor = self.posterior_sampler(posterior)

        # If using botorch based samplers this is an attribute we can access, but we use a less specific interface
        dim_posterior_samples = len(gp_samples.shape) - len(posterior.mean.shape)  # pyright: ignore[reportAttributeAccessIssue]

        # shape is now: (erd_samples_per_period, n_posterior_samples, n_periods, batch_size)
        surrogate_responses = self.sample_surrogate(
            gp_samples,
            self.erd_samples_per_period,
            base_sample_broadcast_dims=list(range(dim_posterior_samples))
            if self.shared_surrogate_base_samples
            else None,
        )

        # For each time period, find the max response
        # Shape is now:  (erd_samples_per_period, n_posterior_samples, n_periods)
        extreme_responses = surrogate_responses.max(dim=-1).values

        return extreme_responses



"""Additional helpers."""

# %%
import json
from pathlib import Path
from typing import Any


def append_to_json(obj: Any, output_file: Path) -> None:  # noqa: ANN401
    """Appends a json object to a file containing a list of objects.

    object: Json serialisable object to append.
    output_file: The file to append results to. If does not exist, it will be created.
    """
    # Initialise the file if it doesn't exist
    if not output_file.exists():
        # Create parent dir if required
        output_file.parent.mkdir(parents=True, exist_ok=True)
        with output_file.open("w") as fp:
            json.dump([], fp)

    with output_file.open("r") as fp:
        existing_results = json.load(fp)

    existing_results.append(obj)

    with output_file.open("w") as fp:
        json.dump(existing_results, fp)



"""Dataset that return importance sample information in the form (data, importance_weight).

Dev details:
    Datasets can return a variety of things when ``__get_item__`` is called, for example:
    - `Tuples <https://pytorch.org/tutorials/beginner/basics/data_tutorial.html#getitem>`_
    - `Dicts <https://pytorch.org/tutorials/beginner/data_loading_tutorial.html#dataset-class>`_

    The dataloader will respect objects like ``dict`` and ``tuple``, and will convert float/int/list/numpy content into
    tensors as it is assumed to be data.
    - By default done by ``collate_fn`` arg of Dataloader.
    - For defaults see ``torch.utils.data._utils.collate.default_collate``

    While this is straight forward to implement, typing can be a challenge (``torch.utils.data.Dataset`` provides some
    guidances, but appears they found it challenging too).

Todo:
    - Revisit the typing and the implications of covariate/contravariant etc.
"""

# pyright: reportUnnecessaryTypeIgnoreComment=false
# %%
from typing import Any, TypeVar

import numpy as np
import torch
from torch.utils.data import Dataset, StackDataset
from torch.utils.data.dataset import T_co, T_tuple

T_np_interface = TypeVar("T_np_interface", np.ndarray[tuple[int, ...], Any], torch.Tensor)


class ImportanceIndexWrapper(Dataset[tuple[T_np_interface, float]]):
    """Wraps an existing dataset, returning a column/index of the item as the importance weight."""

    def __init__(self, dataset: Dataset[T_np_interface], importance_idx: int) -> None:
        """Wrap an existing dataset, returning part of the item as the importance weight.

        Args:
            dataset: A dataset, where one of the columns should be used as an importance weight.
                - Note: Currently only datasets that return numpy or torch tensors are supported
            importance_idx: the column index containing the importance weights.

        Todo:
            - Generalise this to deal with other types of dataset output (list, numpy etc).
        """
        self.dataset: Dataset[T_np_interface] = dataset
        self.importance_idx: int = importance_idx

        datapoint = self.dataset[0]

        self.mask = torch.ones(datapoint.shape, dtype=torch.bool)
        self.mask[importance_idx] = False

    def __len__(self) -> int:  # noqa: D105
        return len(self.dataset)  # type: ignore[arg-type]

    def __getitem__(self, index: int) -> tuple[T_np_interface, float]:
        """Gets the data and importance weight at the requested index in the dataset.

        Args:
            index: what index in the dataset should be returned

        Returns:
            A tuple where:
            - Element 0: (tensor) the datapoint
            - Element 1: Float representing the importance weight
        """
        datapoint = self.dataset[index]

        # datapoint with importance column, importance weight
        return datapoint[self.mask], float(datapoint[self.importance_idx])  # pyright:  ignore[reportReturnType]


class ImportanceAddedWrapper(StackDataset[T_tuple]):  # pyright: ignore[reportMissingTypeArgument]
    "Thin wrapper makes the method for creating the dataset more explicit, and ensure order and type of output."

    def __init__(self, data_dataset: Dataset[T_co], importance_dataset: Dataset[T_co]) -> None:
        """Ensures the StackDataset is created as a tuple regardless of the way args are passed."""
        super().__init__(data_dataset, importance_dataset)



"""Test for gp_bruteforce."""

# %%
from collections.abc import Callable

import matplotlib.pyplot as plt
import pytest
import torch
from botorch.models import SingleTaskGP
from botorch.models.deterministic import GenericDeterministicModel, PosteriorMeanModel
from botorch.models.ensemble import EnsembleModel
from botorch.sampling import IIDNormalSampler
from botorch.sampling.index_sampler import IndexSampler
from tests.qoi.utils import dummy_posterior_mean

from axtreme.plotting.gp_fit import plot_1d_model
from axtreme.qoi.gp_bruteforce import GPBruteForce
from axtreme.sampling import NormalIndependentSampler, PosteriorSampler, UTSampler


class TestSampleSurrogate:
    """Sammple_surrogate: What do we want to test:
    - works with just a single dimension
        - 1 sample
    - works with multiple dimensions (liek the one we will probably use in the problem)
        - One or 2 samples
    - can batch:
        - actual example we intend to use
    """

    def test_single_param(self):
        """Check that the sample_surrogate works with non-batched params and a single sample.

        Because scale is 0, the sample should be equal to the loc.
        """
        loc = torch.tensor([1.0])
        scale = torch.tensor([0.0])

        params = torch.cat([loc, scale], dim=-1)

        sample = GPBruteForce.sample_surrogate(params)

        assert sample == pytest.approx(1.0, abs=1e-5)

    def test_batched_params(self):
        """Check that the sample_surrogate works with batched params and a single sample.

        Because scale is 0, the sample should be equal to the loc. Dimensionality is chosen to mimic what this function
        is commonly used for.
        """
        # Shape represents (n_posterior_samples = 2, n_periods = 3, batch_size = 5, 1)
        # fmt:off
        loc = torch.tensor([
            [[[ 0.],  [1.], [ 2.], [ 3.],  [4.]],
            [[ 5.],  [6.], [ 7.], [ 8.],  [9.]],
            [[10.], [11.], [12.], [13.], [14.]]],

            [[[15.], [16.], [17.], [18.], [19.]],
            [[20.], [21.], [22.], [23.], [24.]],
            [[25.], [26.], [27.], [28.], [29.]]]])
        # fmt: on
        scale = torch.zeros_like(loc) * 0.0

        # Shape represents (n_posterior_samples = 2, n_periods = 3, batch_size = 5, p = 2)
        params = torch.cat([loc, scale], dim=-1)

        sample = GPBruteForce.sample_surrogate(params)

        # shape should be ( number of samples, *b)
        assert sample.shape == torch.Size([1, 2, 3, 5])

        sample1 = sample[0]
        # because scale is 0, the sample should be equal to the loc (need to make the dimension right)
        expected_value = loc.squeeze(-1)
        torch.testing.assert_close(sample1, expected_value, atol=1e-5, rtol=0)

    def test_batched_params_multiple_samples(self):
        """Exactly the same as test__batched_params, but 2 samples are drawn.

        Because scale is 0, the sample should be equal to the loc, and both samples should be the same
        """
        # Shape represents (n_posterior_samples = 2, n_periods = 3, batch_size = 5, 1)
        # fmt:off
        loc = torch.tensor([
            [[[ 0.],  [1.], [ 2.], [ 3.],  [4.]],
            [[ 5.],  [6.], [ 7.], [ 8.],  [9.]],
            [[10.], [11.], [12.], [13.], [14.]]],

            [[[15.], [16.], [17.], [18.], [19.]],
            [[20.], [21.], [22.], [23.], [24.]],
            [[25.], [26.], [27.], [28.], [29.]]]])
        # fmt: on
        scale = torch.zeros_like(loc) * 0.0

        # Shape represents (n_posterior_samples = 2, n_periods = 3, batch_size = 5, p = 2)
        params = torch.cat([loc, scale], dim=-1)

        sample = GPBruteForce.sample_surrogate(params, n_samples=2)

        # shape should be ( number of samples, *b)
        assert sample.shape == torch.Size([2, 2, 3, 5])

        # because scale is 0, the sample should be equal to the loc (need to make the dimension right)
        expected_value = loc.squeeze(-1)
        torch.testing.assert_close(sample[0], expected_value, atol=1e-5, rtol=0)
        torch.testing.assert_close(sample[1], expected_value, atol=1e-5, rtol=0)

    # NOTE: strictly speaking this is non-deterministic, but it should never cause issues.
    def test_shared_base_samples(self):
        """When base samples are shared, the same quantil should be used across the batching demension.

        The following example is close to the usecase for this. Both posteriors should get the same base samples.

        We check this by ensuring corresponsing sample in the two postiors show the same quantile. We use the same scale
        across both posteriors, so we test this by checking they have the same offset from their mean (given by loc).

        Because scale is 0, the sample should be equal to the loc, and both samples should be the same
        """
        # Shape represents (n_posterior_samples = 2, n_periods = 3, batch_size = 5, 1)
        # fmt:off
        loc = torch.tensor([
            [[[ 0.],  [1.], [ 2.], [ 3.],  [4.]],
            [[ 5.],  [6.], [ 7.], [ 8.],  [9.]],
            [[10.], [11.], [12.], [13.], [14.]]],

            [[[15.], [16.], [17.], [18.], [19.]],
            [[20.], [21.], [22.], [23.], [24.]],
            [[25.], [26.], [27.], [28.], [29.]]]])
        # fmt: on
        scale = torch.ones_like(loc) * 1.0

        # Shape represents (n_posterior_samples = 2, n_periods = 3, batch_size = 5, p = 2)
        params = torch.cat([loc, scale], dim=-1)

        sample = GPBruteForce.sample_surrogate(params, n_samples=1, base_sample_broadcast_dims=[0])

        # shape should be ( number of samples, *b)
        assert sample.shape == torch.Size([1, 2, 3, 5])

        # remove the excess dimensions from each
        sample = sample.squeeze(0)
        loc = loc.squeeze(-1)
        # remove the sampling dimension and subtract the element wise mean (using loc)
        offset_posterior1 = sample[0] - loc[0]
        offset_posterior2 = sample[1] - loc[1]

        torch.testing.assert_close(offset_posterior1, offset_posterior2, atol=1e-5, rtol=0)

        # NOTE: Samples should only be share on the n_posterior sample dim. Across other dims things should be unique.
        assert len(offset_posterior1.unique()) == offset_posterior1.numel()


class TestProcessBatch:
    """There is no a large amount of out own code to cover here.

    Integration tests:
    - Check it works with sample_surrogate

    Todo:
        - test with posterior transforms
        - test with posterior the have sample shape that is multidim.
    """

    @pytest.mark.integration
    def test__process_batch(
        self, gp_passthrough_1p: GenericDeterministicModel, gp_passthrough_1p_sampler: IndexSampler
    ):
        """There is a low amount of our own code in"""
        qoi = GPBruteForce(
            env_iterable=[],  # Shouldn't be used in this test
            posterior_sampler=gp_passthrough_1p_sampler,
        )

        # fmt:off
        # Shape is: (n_periods = 3, n_samples_per_period = 5, d = 1)
        env_batch = torch.tensor(
            [
                    [[0.0], [1.0], [2.0], [3.0], [4.0]],
                    [[5.0], [6.0], [7.0], [8.0], [9.0]],
                    [[10.0], [11.0], [12.0], [13.0], [14.0]],

            ]
        )
        # fmt: on

        # Run test
        # expected shape (erd_samples_per_period = 1, n_posterior_samples = 1, n_periods = 3)
        actual_result = qoi._process_batch(env_batch, gp_passthrough_1p)

        # The largest from each row should be selected (in the shape above)
        expected_result = torch.tensor([[[4.0, 9.0, 14.0]]])

        # some relative tollerance allowed as we sample from a very confident gumbel dist
        torch.testing.assert_close(actual_result, expected_result, atol=1e-4, rtol=0)


def env_period_1() -> torch.Tensor:
    """Helper to creates a single period env sample.

    typical shape: (periods, n_samples_per_period, d)

    NOTE: changes here will break test_posterior_samples_erd_samples

    Returns:
    Shape: (1, 5, 1)
    """
    # fmt: off
    env_samples = torch.tensor(
        [
            [[0], [1], [2], [3], [4]],
        ]
    )
    # fmt: on
    return env_samples


def env_period_2() -> torch.Tensor:
    """Helper to creates an env sample with 2 periods.

    typical shape: (periods, n_samples_per_period, d)

    NOTE: changes here will break test_posterior_samples_erd_samples

    Returns:
    Shape: (2, 45 1)
    """
    # fmt: off
    env_samples = torch.tensor(
        [
            [[0], [1], [2], [3], [4]],
            [[9], [8], [7], [6], [5]]
        ]
    )
    # fmt: on
    return env_samples


def batched_env_period() -> list[torch.Tensor]:
    """Batched env data.

    Batching turns;
    env_samples = torch.tensor(
        [
            [[0], [1], [2], [3], [4]], # period1
            [[9], [8], [7], [6], [5]]  # period2
        ]
    )
    Into:
    [
        [
            [[0], [1]],
            [[9], [8]]
        ],
        [
            [[2], [3]],
            [[7], [6]]
        ],
        [
            [[4]],
            [[5]]
        ]
    ]
    """
    # fmt: off
    env_samples = torch.tensor(
        [
            [[0], [1], [2], [3], [4]], # period1
            [[9], [8], [7], [6], [5]]  # period2
        ]
    )
    # fmt: on
    return list(torch.chunk(env_samples, 3, dim=-2))


# TODO(sw 2024-11-22): Now that the function has been refactored this test could be broken up by mocking _process_batch
@pytest.mark.parametrize(
    "env_iterable, n_posterior_samples, erd_samples_per_period, expected_output",
    [
        pytest.param([env_period_1()], 1, 1, torch.tensor([[4]]), id="single_everything"),
        pytest.param([env_period_2()], 1, 1, torch.tensor([[4, 9]]), id="n_period_2"),
        pytest.param([env_period_1()], 2, 1, torch.tensor([[4], [4]]), id="n_posterior_samples_2"),
        pytest.param([env_period_1()], 1, 2, torch.tensor([[4, 4]]), id="erd_samples_2"),
        # Should produce the exact same results as n_period_2 test above
        pytest.param(batched_env_period(), 1, 1, torch.tensor([[4, 9]]), id="batched_env"),
    ],
)
def test_posterior_samples_erd_samples(
    env_iterable: list[torch.Tensor],
    n_posterior_samples: int,
    erd_samples_per_period: int,
    expected_output: torch.Tensor,
    gp_passthrough_1p: GenericDeterministicModel,
):
    """Checks the ERD samples for different configurations of GPBruteForce.

    By fixed the model, posterior samples, and gumbel samples produced, we can determinsitically trace input values
    through to their expect value in the erd samples. This is done by using `dummy_posterior_mean` .

    Specifically this tests:
    - Single sample of all params returns the 1 ERD sample.
    - multiple periods: each one generates and ERD sample.
    - multiple posterior samples: each generate a set of ERD samples
    - multiple erd_samples_per_period: multiple ERD samples are generated per period
    - Batching: doesn't effect the ERD generated

    Args:
        env_iterable: can be of the following shape;
            - non-batched shape: (1, periods, n_samples_per_period, d), where d=1
            -  batched shape: (n_batches, periods, n_samples_per_period//n_batches, d), d=1
        n_posterior_samples: number of posterior samples taken
        erd_samples_per_period: number of erd samples taken
        expected_output: output of shape (n_posterior_samples, total_erd_samples) expected.
    """
    # Create GP like object around dummy_posterior_mean
    model = gp_passthrough_1p
    # Sampler that belongs with DeterministicModels
    sampler = IndexSampler(torch.Size([n_posterior_samples]))

    qoi = GPBruteForce(env_iterable, posterior_sampler=sampler, erd_samples_per_period=erd_samples_per_period)

    # produces output of shape (n_posterior_samples, n_periods * erd_samples_per_period)
    ests = qoi.posterior_samples_erd_samples(model)

    # using atol = 1e-5 covers the entire support of Gumbel(loc = x, scale = 1e-6)
    torch.testing.assert_close(ests, expected_output, atol=1e-5, rtol=0, check_dtype=False)


# %%
def precalculate_2_posterior_samples(X: torch.Tensor) -> torch.Tensor:  # noqa: N803
    """Creates 2 posterior samples, where sample2 = sample1 + 10.

    Using this in conjuction with:
        - GenericDeterministicModel: to create an Ensemble model
        - IndexSampler with base_samples = torch.tensor([0,1])

    Args:
    X: typically (n_periods, n, d)

    Return:
    (*b,s,n,t):
        - *b batch dimension
        - s: number of ensembels (e.g number of samples at point x)
        - n: number of input points
        - t: number of targets
    """
    posterior1 = dummy_posterior_mean(X)
    posterior2 = dummy_posterior_mean(X)
    # increase the loc value
    posterior2[..., 0] += 10

    # this is shape (s, n_periods, n, t)
    output = torch.stack([posterior1, posterior2])
    # Swap it to be of the required form *b,s,n,t)
    return torch.swapaxes(output, 0, 1)


class GenericEnsembleModel(EnsembleModel):
    r"""A generic ensembel model constructed from a callable."""

    def __init__(self, f: Callable[[torch.Tensor], torch.Tensor]) -> None:
        """Create an Ensemble model from an underling function.

        Args:
            f: A callable mapping a `batch_shape x n x d`-dim input tensor `x`
                to a `batch_shape x s x n x m`-dimensional output tensor (the
                outcome dimension `m` must be explicit, even if `m=1`). `s` is the number
        """
        super().__init__()
        self._f = f

    def forward(self, X: torch.Tensor) -> torch.Tensor:  # noqa: N803 # for consitence with botorch interface
        r"""Compute the (deterministic) model output at X.

        Args:
            X: A `batch_shape x n x d`-dim input tensor `X`.

        Returns:
            A `batch_shape x n x m`-dimensional output tensor.
        """
        return self._f(X)


def test_posterior_samples_erd_samples_grouped_by_posterior():
    """Ensure that ERD samples are listed under the posterior that generated them.

    The mocked gp and posterior combination means that posterior samples produced for loc will be:
    [
        # posterior sample 2
        [
            [[0], [1], [2], [3], [4]],
            [[9], [8], [7], [6], [5]]
        ],
        # posterior sample 2
        [
            [[10], [11], [12], [13], [14]],
            [[19], [18], [17], [16], [15]]
        ]
    ]
    Beause the scale is ~0, the Gumbel samples will be the same as the loc values.
    """
    # fmt: off
    env_data = torch.tensor(
        [
            [[0], [1], [2], [3], [4]],
            [[9], [8], [7], [6], [5]]
        ]
    )
    # fmt: on

    # NOTE: The combination of precalculate_2_posterior_samples, GenericEnsembleModel, IndexSampler has been tested
    # manually to check it produces the required output
    model = GenericEnsembleModel(precalculate_2_posterior_samples)

    # Chose the order that these are sampled in
    sampler = IndexSampler(torch.Size([2]))
    sampler.base_samples = torch.tensor([0, 1])

    qoi = GPBruteForce([env_data], posterior_sampler=sampler)

    # produces output of shape (n_posterior_samples, n_periods * erd_samples_per_period)
    ests = qoi.posterior_samples_erd_samples(model)

    # fmt: off
    expected_output = torch.tensor(
        [
            [4,9],
            [14,19]
        ]
    )
    # fmt: on

    # using atol = 1e-5 covers the entire support of Gumbel(loc = x, scale = 1e-6)
    torch.testing.assert_close(ests, expected_output, atol=1e-5, rtol=0, check_dtype=False)


def test_grad():
    # env_iterable and posterior sampler are not used in this test
    with pytest.raises(NotImplementedError):
        _ = GPBruteForce(env_iterable=[], posterior_sampler=IndexSampler(torch.Size([1])), no_grad=False)


def test_gpu():
    # env_iterable and posterior sampler are not used in this test
    with pytest.raises(NotImplementedError):
        _ = GPBruteForce(env_iterable=[], posterior_sampler=IndexSampler(torch.Size([1])), device=torch.device("cuda"))


# NOTE: This functionaltiy is important for GPBruteForce, so the test is put here, but could in in a number of places.
@pytest.mark.integration
@pytest.mark.parametrize(
    "sampler",
    [
        pytest.param(NormalIndependentSampler(torch.Size([5]), seed=7), id="NormalIndependentSampler"),
        pytest.param(UTSampler(), id="UTSampler"),
        pytest.param(
            IIDNormalSampler(torch.Size([5]), seed=7),
            marks=pytest.mark.xfail(reason="Demonstration of non-batchable posterior"),
            id="IIDNormalSampler (from botorch)",
        ),
    ],
)
def test_posterior_samples_are_equivalent_when_batched(
    sampler: PosteriorSampler,
    model_singletaskgp_d1_t1: SingleTaskGP,
    *,
    visual_inspect: bool = False,
):
    """Tests if posterior samplers are invariant to batching on input data as is required for GPBruteForce.

    Often we work with periods of env data that are too large to predict a single posterior for. In this case it is
    useful to batch the data, and run the GP on each batch.One shortcoming is that the covariance between items in
    different batches is lost. Posterior samplers that make use of the covariance between different points will produce
    different samples depending on if the data is batched or not. This can be problematic for the following reasons:
    - Ideally the result should be invariant to the batch dimension choosen.
    - In the batch case the posterior samples will be discontinous, which can mean samples no longer reflect behviour of
      the underling function being modeled.

    The following compares the posterior generate on non-batched and batched data, and checks if they are the same.
    `visual_inspect` can be used when running this function manually to visually check the results.
    """
    model = model_singletaskgp_d1_t1

    #### None of following
    env_data = torch.linspace(0, 1, 100).reshape(-1, 1)

    # not batched process
    env_iterable = [env_data]

    for data in env_iterable:
        with torch.no_grad():
            posterior = model.posterior(data)

        samples = sampler(posterior)
        samples = samples.squeeze(-1)  # flatten the t=1 dimension

    if visual_inspect:
        ax = plot_1d_model(model)
        for sample in samples:
            _ = ax.plot(env_data.flatten(), sample, color="grey")
        plt.show()

    # Batch processing
    batch_env_iterable = torch.chunk(env_data, 3)

    batches = []
    for data in batch_env_iterable:
        with torch.no_grad():
            posterior = model.posterior(data)
        batch_samples = sampler(posterior)
        batch_samples = batch_samples.squeeze()
        batches.append(batch_samples)

    if visual_inspect:
        ax = plot_1d_model(model)
        colour = ["r", "g", "b"]
        for batch_x, batch_samples, c in zip(batch_env_iterable, batches, colour, strict=True):
            for sample in batch_samples:
                _ = ax.plot(batch_x.flatten(), sample, color=c)
        plt.show()

    # Check the batch sample and the full samples are the same once batches are combined
    # batches have dimensions (n_posterior_samples, batch_size).
    # Combine the batches so get back to shape (n_posterior_samples, env_samples)
    joined_batches = torch.cat(batches, dim=-1)
    torch.testing.assert_close(samples, joined_batches)


def test_shared_surrogate_base_samples_no_gp_uncertainty(model_singletaskgp_d1_t2: SingleTaskGP):
    """With shared surrogate bases sample and not GP uncertainty, all qoi outputs should be identical.

    When there is no GP uncertainty, all posterior samples are identical. This means they output the exact same
    distributions at each x point. When base samples are shared across posteriors, the exact sample point/quantile is
    sampled from each of the identical distributions. As a result all QoI posteriors should produce the same result.
    """

    model = PosteriorMeanModel(model_singletaskgp_d1_t2)
    sampler = IndexSampler(torch.Size([10]))

    # (n_periods, period_len, d)
    env_data = torch.rand(3, 5, 1)

    qoi = GPBruteForce([env_data], posterior_sampler=sampler, shared_surrogate_base_samples=True)

    results = qoi(model)

    assert torch.all(results == results[0])



"""helpers for working with distribution."""

import torch
from torch.distributions import Distribution, MixtureSameFamily


def index_batch_dist(dist: Distribution, index: tuple[slice | int, ...]) -> Distribution:
    """Applies indexing to a MixtureSameFamily object along the batch dimension.

    Args:
        dist: The distribution to be indexed (only along the batch dimensions)
        index: The index/slice to be applied. e.g `(slice(1,4), Ellipsis)` is equivalent to [1:4,...].
            Slices larger than the batch dimension will cause index error.

    Returns:
        A "veiw" of the underling distribution. This is a new object, but we call it a veiw as it is built on a view of
        the underling data.

    Note:
        The returned distribution is built on a view of the underling data. The follows the behaviour of slicing tensors
        in pytorch as detailed `here <https://discuss.pytorch.org/t/does-indexing-a-tensor-return-a-copy-of-it/164905>`_.
        As such gradients etc are connected.
    """
    # Mixture distribution require special handling because they have multiple distributions internally.
    if isinstance(dist, MixtureSameFamily):
        return index_batch_mixture_dist(dist, index)

    # NOTE: unclear if all dists have a populated arg_constraints. If not, this will fail and we will be alerted to
    # the issues
    params = {param: getattr(dist, param)[index] for param in dist.arg_constraints}
    return dist.__class__(**params)


def index_batch_mixture_dist(dist: MixtureSameFamily, index: tuple[slice | int, ...]) -> MixtureSameFamily:
    """Applies indexing to a MixtureSameFamily object along the batch dimension.

    Args:
        dist: The distribution to be indexed (only along the batch dimensions)
        index: The index/slice to be applied. e.g `(slice(1,4), Ellipsis)` is equivalent to [1:4,...].
            Slices larger than the batch dimension will cause index error.

    Returns:
        A "veiw" of the underling distribution. This is a new object, but we call it a veiw as it is built on a view of
        the underling data.

    Note:
        The returned distribution is built on a view of the underling data. The follows the behaviour of slicing tensors
        in pytorch as detailed `here <https://discuss.pytorch.org/t/does-indexing-a-tensor-return-a-copy-of-it/164905>`_.
        As such gradients etc are connected.
    """
    if len(index) > len(dist.batch_shape):
        raise IndexError(f"too many indices for tensor of dimension {dist.batch_shape}")

    # NOTE: unclear if all dists have a populated arg_constraints. If not, this will fail and we will be alerted to
    # the issues

    comp_slice = {
        param: getattr(dist.component_distribution, param)[index]
        for param in dist.component_distribution.arg_constraints
    }
    # both probs and logits are automatically populated as soon as one is provided. We an pick either.
    mix_slice = {"probs": dist.mixture_distribution.probs[index]}

    comp_new = dist.component_distribution.__class__(**comp_slice)
    mix_new = dist.mixture_distribution.__class__(**mix_slice)

    return dist.__class__(mix_new, comp_new)


def dist_dtype(dist: Distribution) -> torch.dtype:
    """Return the dtype the distribution calculates values in.

    Parameters may be of different tpyes. It appears the distribution defaults to the largest dtype.

    Args:
        dist: the distribution to find the dtype of.

    Returns:
        dtype the ditribution returns result in.
    """
    # All distribution support this attribute. This is the easiest way to identify the dtype.
    return dist.mean.dtype



"""These items typically are part of creating fixture in conftest.py, but are also imported directly in the test.

Placed here as pytest compains when importing from conftest directly.
"""

import torch


def dummy_posterior_mean(x: torch.Tensor) -> torch.Tensor:
    """This is a helper function to be used with GenericDeterministicModel to create models with deterministic output.

    This can be created with the following:
    > model = GenericDeterministicModel(dummpy_posterior_mean, num_outputs=2)

    The behaviour of this function is to create:
        - `loc` is set equal to the `env_value`
        - `scale` is set to 1e-6

    Useful because the result of sampling from the subsequent gumbel distibution is effectively deterministic.

    Args:
    x: (*b,d=1): a batch of env data, where each env datapoint has dim=1

    Return:
    (b*, 2): where the last dimension is [loc, scale]
    """
    scale = torch.ones_like(x) * 1e-6
    return torch.concat([x, scale], dim=-1)
